import re
from collections import deque
from queue import LifoQueue

import networkx as nx
import numpy as np
from networkx.algorithms.dag import is_directed_acyclic_graph
from networkx.algorithms.shortest_paths.generic import shortest_path

from dowhy.utils.ordered_set import OrderedSet


def adjacency_matrix_to_adjacency_list(adjacency_matrix, labels=None):
    """
    Convert the adjacency matrix of a graph to an adjacency list.

    :param adjacency_matrix: A numpy array representing the graph adjacency matrix.
    :param labels: List of labels.
    :returns: Adjacency list as a dictionary.
    """

    adjlist = dict()
    if labels is None:
        labels = [str(i + 1) for i in range(adjacency_matrix.shape[0])]
    for i in range(adjacency_matrix.shape[0]):
        adjlist[labels[i]] = list()
        for j in range(adjacency_matrix.shape[1]):
            if adjacency_matrix[i, j] != 0:
                adjlist[labels[i]].append(labels[j])
    return adjlist


def adjacency_matrix_to_graph(adjacency_matrix, labels=None):
    """
    Convert a given graph adjacency matrix to DOT format.

    :param adjacency_matrix: A numpy array representing the graph adjacency matrix.
    :param labels: List of labels.
    :returns: Graph in DOT format.
    """
    # Only consider edges have absolute edge weight > 0.01
    idx = np.abs(adjacency_matrix) > 0.01
    dirs = np.where(idx)
    import graphviz

    d = graphviz.Digraph(engine="dot")
    names = labels if labels else [f"x{i}" for i in range(len(adjacency_matrix))]
    for name in names:
        d.node(name)
    for to, from_, coef in zip(dirs[0], dirs[1], adjacency_matrix[idx]):
        d.edge(names[from_], names[to], label=str(coef))
    return d


def str_to_dot(string):
    """
    Converts input string from graphviz library to valid DOT graph format.

    :param string: Graph in DOT format.
    :returns: DOT string converted to a suitable format for the DoWhy library.
    """
    graph = string.strip().replace("\n", ";").replace("\t", "")
    graph = graph[:9] + graph[10:-2] + graph[-1]  # Removing unnecessary characters from string
    return graph


def find_ancestor(node_set, node_names, adjacency_matrix, node2idx, idx2node):
    """
    Finds ancestors of a given set of nodes in a given graph.

    :param node_set: Set of nodes whos ancestors must be obtained.
    :param node_names: Name of all nodes in the graph.
    :param adjacency_matrix: Graph adjacency matrix.
    :param node2idx: A dictionary mapping node names to their row or column index in the adjacency matrix.
    :param idx2node: A dictionary mapping the row or column indices in the adjacency matrix to the corresponding node names.
    :returns: OrderedSet containing ancestors of all nodes in the node_set.
    """

    def find_ancestor_help(node_name, node_names, adjacency_matrix, node2idx, idx2node):
        ancestors = OrderedSet()
        nodes_to_visit = LifoQueue(maxsize=len(node_names))
        nodes_to_visit.put(node2idx[node_name])
        while not nodes_to_visit.empty():
            child = nodes_to_visit.get()
            ancestors.add(idx2node[child])
            for i in range(len(node_names)):
                if (
                    idx2node[i] not in ancestors and adjacency_matrix[i, child] == 1
                ):  # For edge a->b, a is along height and b is along width of adjacency matrix
                    nodes_to_visit.put(i)
        return ancestors

    ancestors = OrderedSet()
    for node_name in node_set.get_all():
        ancestors = ancestors.union(find_ancestor_help(node_name, node_names, adjacency_matrix, node2idx, idx2node))
    return ancestors


def induced_graph(node_set, adjacency_matrix, node2idx):
    """
    To obtain the induced graph corresponding to a subset of nodes.

    :param node_set: Set of nodes whos ancestors must be obtained.
    :param adjacency_matrix: Graph adjacency matrix.
    :param node2idx: A dictionary mapping node names to their row or column index in the adjacency matrix.
    :returns: Numpy array representing the adjacency matrix of the induced graph.
    """
    node_idx_list = [node2idx[node] for node in node_set]
    node_idx_list.sort()
    adjacency_matrix_induced = adjacency_matrix.copy()
    adjacency_matrix_induced = adjacency_matrix_induced[node_idx_list]
    adjacency_matrix_induced = adjacency_matrix_induced[:, node_idx_list]
    return adjacency_matrix_induced


def find_c_components(adjacency_matrix, node_set, idx2node):
    """
    Obtain C-components in a graph.

    :param adjacency_matrix: Graph adjacency matrix.
    :param node_set: Set of nodes whos ancestors must be obtained.
    :param idx2node: A dictionary mapping the row or column indices in the adjacency matrix to the corresponding node names.
    :returns: List of C-components in the graph.
    """
    num_nodes = len(node_set)
    adj_matrix = adjacency_matrix.copy()
    adjacency_list = [[] for _ in range(num_nodes)]

    # Modify graph such that it only contains bidirected edges
    for h in range(0, num_nodes - 1):
        for w in range(h + 1, num_nodes):
            if adjacency_matrix[h, w] == 1 and adjacency_matrix[w, h] == 1:
                adjacency_list[h].append(w)
                adjacency_list[w].append(h)
            else:
                adj_matrix[h, w] = 0
                adj_matrix[w, h] = 0

    # Find c components by finding connected components on the undirected graph
    visited = [False for _ in range(num_nodes)]

    def dfs(node_idx, component):
        visited[node_idx] = True
        component.add(idx2node[node_idx])
        for neighbour in adjacency_list[node_idx]:
            if visited[neighbour] == False:
                dfs(neighbour, component)

    c_components = []
    for i in range(num_nodes):
        if visited[i] == False:
            component = OrderedSet()
            dfs(i, component)
            c_components.append(component)

    return c_components


def daggity_to_dot(daggity_string):
    """
    Converts the input daggity_string to valid DOT graph format.

    :param daggity_string: Output graph from Daggity site
    :returns: DOT string
    """
    graph = re.sub(r"\n", "; ", daggity_string)
    graph = re.sub(r"^dag ", "digraph ", graph)
    graph = re.sub("{;", "{", graph)
    graph = re.sub("};", "}", graph)
    graph = re.sub("outcome,*,", "", graph)
    graph = re.sub("adjusted,*", "", graph)
    graph = re.sub("exposure,*", "", graph)
    graph = re.sub("latent,*", 'observed="no",', graph)
    graph = re.sub(",]", "]", graph)
    return graph


def get_simple_ordered_tree(n):
    """
    Generates a simple-ordered tree. The tree is just a
    directed acyclic graph of n nodes with the structure
    0 --> 1 --> .... --> n.
    """
    g = nx.DiGraph()

    for i in range(n):
        g.add_node(i)

    for i in range(n - 1):
        g.add_edges_from([(i, i + 1, {})])
    return g


def is_connected(g):
    """
    Checks if a the directed acyclic graph is connected.
    """
    u = convert_to_undirected_graph(g)
    return nx.is_connected(u)


def convert_to_undirected_graph(g):
    u = nx.Graph()
    for n in g.nodes:
        u.add_node(n)
    for e in g.edges:
        u.add_edges_from([(e[0], e[1], {})])
    return u


def get_random_node_pair(n):
    """
    Randomly generates a pair of nodes.
    """
    i = np.random.randint(0, n)
    j = i
    while j == i:
        j = np.random.randint(0, n)
    return i, j


def find_predecessor(i, j, g):
    """
    Finds a predecessor, k, in the path between two nodes, i and j,
    in the graph, g.
    """
    parents = list(g.predecessors(j))
    u = convert_to_undirected_graph(g)
    for pa in parents:
        try:
            path = shortest_path(u, pa, i)
            return pa
        except:
            pass
    return None


def del_edge(i, j, g):
    """
    Deletes the edge i --> j in the graph, g. The edge is only
    deleted if this removal does NOT cause the graph to be
    disconnected.
    """
    if g.has_edge(i, j) is True:
        g.remove_edge(i, j)

        if is_connected(g) is False:
            g.add_edges_from([(i, j, {})])


def add_edge(i, j, g):
    """
    Adds an edge i --> j to the graph, g. The edge is only
    added if this addition does NOT cause the graph to have
    cycles.
    """
    g.add_edges_from([(i, j, {})])
    if is_directed_acyclic_graph(g) is False:
        g.remove_edge(i, j)



import logging
from copy import deepcopy
from typing import Any, Dict, List, Optional, Tuple

import networkx as nx
import numpy as np
import pandas as pd
from matplotlib import pyplot
from networkx.drawing import nx_pydot

_logger = logging.getLogger(__name__)


def plot(
    causal_graph: nx.Graph,
    causal_strengths: Optional[Dict[Tuple[Any, Any], float]] = None,
    filename: Optional[str] = None,
    display_plot: bool = True,
    figure_size: Optional[List[int]] = None,
    **kwargs,
) -> None:
    """Convenience function to plot causal graphs. This function uses different backends based on what's
    available on the system. The best result is achieved when using Graphviz as the backend. This requires both
    the Python pygraphviz package (``pip install pygraphviz``) and the shared system library (e.g. ``brew install
    graphviz`` or ``apt-get install graphviz``). When graphviz is not available, it will fall back to the
    networkx backend.

    :param causal_graph: The graph to be plotted
    :param causal_strengths: An optional dictionary with Edge -> float entries.
    :param filename: An optional filename if the output should be plotted into a file.
    :param display_plot: Optionally specify if the plot should be displayed or not (default to True).
    :param figure_size: A tuple to define the width and height (as a tuple) of the pyplot. This is used to parameter to
                        modify pyplot's 'figure.figsize' parameter. If None is given, the current/default value is used.
    :param kwargs: Remaining parameters will be passed through to the backend verbatim.

    **Example usage**::

    >>> plot(nx.DiGraph([('X', 'Y')])) # plots X -> Y
    >>> plot(nx.DiGraph([('X', 'Y')]), causal_strengths={('X', 'Y'): 0.43}) # annotates arrow with 0.43
    """
    try:
        from dowhy.gcm.util.pygraphviz import _plot_causal_graph_graphviz

        try:
            _plot_causal_graph_graphviz(
                causal_graph,
                causal_strengths=causal_strengths,
                filename=filename,
                display_plot=display_plot,
                figure_size=figure_size,
                **kwargs,
            )
        except Exception as error:
            _logger.info(
                "There was an error when trying to plot the graph via graphviz, falling back to networkx "
                "plotting. If graphviz is not installed, consider installing it for better looking plots. The"
                " error is:" + str(error)
            )
            _plot_causal_graph_networkx(
                causal_graph,
                causal_strengths=causal_strengths,
                filename=filename,
                display_plot=display_plot,
                figure_size=figure_size,
                **kwargs,
            )

    except ImportError:
        _logger.info(
            "Pygraphviz installation not found, falling back to networkx plotting. "
            "For better looking plots, consider installing pygraphviz. Note This requires both the Python "
            "pygraphviz package (``pip install pygraphviz``) and the shared system library (e.g. "
            "``brew install graphviz`` or ``apt-get install graphviz``)"
        )
        _plot_causal_graph_networkx(
            causal_graph,
            causal_strengths=causal_strengths,
            filename=filename,
            display_plot=display_plot,
            figure_size=figure_size,
            **kwargs,
        )


def plot_adjacency_matrix(
    adjacency_matrix: pd.DataFrame, is_directed: bool, filename: Optional[str] = None, display_plot: bool = True
) -> None:
    plot(
        nx.from_pandas_adjacency(adjacency_matrix, nx.DiGraph() if is_directed else nx.Graph()),
        display_plot=display_plot,
        filename=filename,
    )


def _plot_causal_graph_networkx(
    causal_graph: nx.Graph,
    pydot_layout_prog: Optional[str] = None,
    causal_strengths: Optional[Dict[Tuple[Any, Any], float]] = None,
    filename: Optional[str] = None,
    display_plot: bool = True,
    label_wrap_length: int = 3,
    figure_size: Optional[List[int]] = None,
) -> None:
    if "graph" not in causal_graph.graph:
        causal_graph.graph["graph"] = {"rankdir": "TD"}

    if pydot_layout_prog is not None:
        layout = nx_pydot.pydot_layout(causal_graph, prog=pydot_layout_prog)
    else:
        layout = nx.spring_layout(causal_graph)

    if causal_strengths is None:
        causal_strengths = {}
    else:
        causal_strengths = deepcopy(causal_strengths)

    max_strength = 0.0
    for (source, target, strength) in causal_graph.edges(data="CAUSAL_STRENGTH", default=1):
        if (source, target) not in causal_strengths:
            causal_strengths[(source, target)] = strength
        max_strength = max(max_strength, abs(causal_strengths[(source, target)]))

    for edge in causal_graph.edges:
        if edge[0] == edge[1]:
            raise ValueError(
                "Node %s has a self-cycle, i.e. a node pointing to itself. Plotting self-cycles is "
                "currently only supported for plots using Graphviz! Consider installing the corresponding"
                "requirements." % edge[0]
            )

    # Wrapping labels if they are too long
    labels = {}
    for node in causal_graph.nodes:
        node_name_splits = str(node).split(" ")
        for i in range(1, len(node_name_splits)):
            if len(node_name_splits[i - 1]) > label_wrap_length:
                node_name_splits[i] = "\n" + node_name_splits[i]
            else:
                node_name_splits[i] = " " + node_name_splits[i]

        labels[node] = "".join(node_name_splits)

    if figure_size is not None:
        org_fig_size = pyplot.rcParams["figure.figsize"]
        pyplot.rcParams["figure.figsize"] = figure_size

    figure = pyplot.figure()

    nx.draw(
        causal_graph,
        pos=layout,
        node_color="lightblue",
        linewidths=0.25,
        labels=labels,
        font_size=8,
        font_weight="bold",
        node_size=2000,
        width=[_calc_arrow_width(causal_strengths[(s, t)], max_strength) for (s, t) in causal_graph.edges()],
    )

    if display_plot:
        pyplot.show()

    if figure_size is not None:
        pyplot.rcParams["figure.figsize"] = org_fig_size

    if filename is not None:
        figure.savefig(filename)


def _calc_arrow_width(strength: float, max_strength: float):
    return 0.2 + 4.0 * float(abs(strength)) / float(max_strength)


def bar_plot(
    values: Dict[str, float],
    uncertainties: Optional[Dict[str, Tuple[float, float]]] = None,
    ylabel: str = "",
    filename: Optional[str] = None,
    display_plot: bool = True,
    figure_size: Optional[List[int]] = None,
    bar_width: float = 0.8,
    xticks: List[str] = None,
    xticks_rotation: int = 90,
) -> None:
    """Convenience function to make a bar plot of the given values with uncertainty bars, if provided. Useful for all
    kinds of attribution results (including confidence intervals).

    :param values: A dictionary where the keys are the labels and the values are the values to be plotted.
    :param uncertainties: A dictionary of attributes to be added to the error bars.
    :param ylabel: The label for the y-axis.
    :param filename: An optional filename if the output should be plotted into a file.
    :param display_plot: Optionally specify if the plot should be displayed or not (default to True).
    :param figure_size: The size of the figure to be plotted.
    :param bar_width: The width of the bars.
    :param xticks: Explicitly specify the labels for the bars on the x-axis.
    :param xticks_rotation: Specify the rotation of the labels on the x-axis.
    """
    if uncertainties is None:
        uncertainties = {node: [values[node], values[node]] for node in values}

    figure, ax = pyplot.subplots(figsize=figure_size)
    ci_plus = [uncertainties[node][1] - values[node] for node in values.keys()]
    ci_minus = [values[node] - uncertainties[node][0] for node in values.keys()]
    yerr = np.array([ci_minus, ci_plus])
    yerr[abs(yerr) < 10**-7] = 0
    pyplot.bar(values.keys(), values.values(), yerr=yerr, ecolor="#1E88E5", color="#ff0d57", width=bar_width)
    pyplot.ylabel(ylabel)
    pyplot.xticks(rotation=xticks_rotation)

    ax.spines["right"].set_visible(False)
    ax.spines["top"].set_visible(False)
    if xticks:
        pyplot.xticks(list(uncertainties.keys()), xticks)

    if display_plot:
        pyplot.show()

    if filename is not None:
        figure.savefig(filename)



from typing import Dict, List, Optional, Set, Union

import networkx as nx
import numpy as np

from dowhy.causal_graph import CausalGraph
from dowhy.utils.api import parse_state
from dowhy.utils.graph_operations import find_ancestor, find_c_components, induced_graph
from dowhy.utils.ordered_set import OrderedSet


class IDExpression:
    """
    Class for storing a causal estimand, as a result of the identification step using the ID algorithm.
    The object stores a list of estimators(self._product) whose porduct must be obtained and a list of variables (self._sum) over which the product must be marginalized.
    """

    def __init__(self):
        self._product = []
        self._sum = []

    def add_product(self, element: Union[Dict, "IDExpression"]):
        """
        Add an estimator to the list of product.

        :param element: Estimator to append to the product list.
        """
        self._product.append(element)

    def add_sum(self, element: Set):
        """
        Add variables to the list.

        :param element: Set of variables to append to the list self._sum.
        """
        for el in element:
            self._sum.append(el)

    def get_val(self, return_type: str):
        """
        Get either the list of estimators (for product) or list of variables (for the marginalization).

        :param return_type: "prod" to return the list of estimators or "sum" to return the list of variables.
        """
        if return_type == "prod":
            return self._product
        elif return_type == "sum":
            return self._sum
        else:
            raise Exception("Provide correct return type.")

    def _print_estimator(self, prefix, estimator: Union[Dict, "IDExpression"] = None, start: bool = False):
        """
        Print the IDExpression object.
        """
        if estimator is None:
            return None

        string = ""
        if isinstance(estimator, IDExpression):
            s = True if len(estimator.get_val(return_type="sum")) > 0 else False
            if s:
                sum_vars = "{" + ",".join(estimator.get_val(return_type="sum")) + "}"
                string += prefix + "Sum over " + sum_vars + ":\n"
                prefix += "\t"
            for expression in estimator.get_val(return_type="prod"):
                add_string = self._print_estimator(prefix, expression)
                if add_string is None:
                    return None
                else:
                    string += add_string
        else:
            outcome_vars = list(estimator["outcome_vars"])
            condition_vars = list(estimator["condition_vars"])
            string += prefix + "Predictor: P(" + ",".join(outcome_vars)
            if len(condition_vars) > 0:
                string += "|" + ",".join(condition_vars)
            string += ")\n"
        if start:
            string = string[:-1]
        return string

    def __str__(self):
        string = self._print_estimator(prefix="", estimator=self, start=True)
        if string is None:
            return "The graph is not identifiable."
        else:
            return string


class IDIdentifier:
    """
    This class is for backwards compatibility with CausalModel
    Will be deprecated in the future in favor of function call id_identify_effect()
    """

    def identify_effect(
        self,
        graph: CausalGraph,
        treatment_name: Union[str, List[str]],
        outcome_name: Union[str, List[str]],
        node_names: Optional[Union[str, List[str]]] = None,
        **kwargs,
    ):
        return identify_effect_id(graph, treatment_name, outcome_name, node_names, **kwargs)


def identify_effect_id(
    graph: CausalGraph,
    treatment_name: Union[str, List[str]],
    outcome_name: Union[str, List[str]],
    node_names: Optional[Union[str, List[str]]] = None,
    **kwargs,
) -> IDExpression:
    """
    Implementation of the ID algorithm.
    Link - https://ftp.cs.ucla.edu/pub/stat_ser/shpitser-thesis.pdf
    The pseudo code has been provided on Pg 40.

    :param treatment_names: OrderedSet comprising names of treatment variables.
    :param outcome_names:OrderedSet comprising names of outcome variables.
    :param node_names: OrderedSet comprising names of all nodes in the graph

    :returns: target estimand, an instance of the IDExpression class.
    """

    if node_names is None:
        node_names = OrderedSet(graph._graph.nodes)

    adjacency_matrix = graph.get_adjacency_matrix()

    try:
        tsort_node_names = OrderedSet(list(nx.topological_sort(graph._graph)))  # topological sorting of graph nodes
    except:
        raise Exception("The graph must be a directed acyclic graph (DAG).")

    return __adjacency_matrix_identify_effect(
        adjacency_matrix,
        OrderedSet(parse_state(treatment_name)),
        OrderedSet(parse_state(outcome_name)),
        tsort_node_names,
        node_names,
    )


def __adjacency_matrix_identify_effect(
    adjacency_matrix: np.matrix,
    treatment_name: OrderedSet,
    outcome_name: OrderedSet,
    tsort_node_names: OrderedSet,
    node_names: OrderedSet = None,
):

    node2idx, idx2node = __idx_node_mapping(node_names)

    # Estimators list for returning after identification
    estimators = IDExpression()

    # Line 1
    # If no action has been taken, the effect on Y is just the marginal of the observational distribution P(v) on Y.
    if len(treatment_name) == 0:
        identifier = IDExpression()
        estimator = {}
        estimator["outcome_vars"] = node_names
        estimator["condition_vars"] = OrderedSet()
        identifier.add_product(estimator)
        identifier.add_sum(node_names.difference(outcome_name))
        estimators.add_product(identifier)
        return estimators

    # Line 2
    # If we are interested in the effect on Y, it is sufficient to restrict our attention on the parts of the model ancestral to Y.
    ancestors = find_ancestor(outcome_name, node_names, adjacency_matrix, node2idx, idx2node)
    if (
        len(node_names.difference(ancestors)) != 0
    ):  # If there are elements which are not the ancestor of the outcome variables
        # Modify list of valid nodes
        treatment_name = treatment_name.intersection(ancestors)
        node_names = node_names.intersection(ancestors)
        adjacency_matrix = induced_graph(node_set=node_names, adjacency_matrix=adjacency_matrix, node2idx=node2idx)
        return __adjacency_matrix_identify_effect(
            treatment_name=treatment_name,
            outcome_name=outcome_name,
            adjacency_matrix=adjacency_matrix,
            tsort_node_names=tsort_node_names,
            node_names=node_names,
        )

    # Line 3 - forces an action on any node where such an action would have no effect on Y – assuming we already acted on X.
    # Modify adjacency matrix to obtain that corresponding to do(X)
    adjacency_matrix_do_x = adjacency_matrix.copy()
    for x in treatment_name:
        x_idx = node2idx[x]
        for i in range(len(node_names)):
            adjacency_matrix_do_x[i, x_idx] = 0
    ancestors = find_ancestor(outcome_name, node_names, adjacency_matrix_do_x, node2idx, idx2node)
    W = node_names.difference(treatment_name).difference(ancestors)
    if len(W) != 0:
        return __adjacency_matrix_identify_effect(
            treatment_name=treatment_name.union(W),
            outcome_name=outcome_name,
            adjacency_matrix=adjacency_matrix,
            tsort_node_names=tsort_node_names,
            node_names=node_names,
        )

    # Line 4 - Decomposes the problem into a set of smaller problems using the key property of C-component factorization of causal models.
    # If the entire graph is a single C-component already, further problem decomposition is impossible, and we must provide base cases.
    # Modify adjacency matrix to remove treatment variables
    node_names_minus_x = node_names.difference(treatment_name)
    node2idx_minus_x, idx2node_minus_x = __idx_node_mapping(node_names_minus_x)
    adjacency_matrix_minus_x = induced_graph(
        node_set=node_names_minus_x, adjacency_matrix=adjacency_matrix, node2idx=node2idx
    )
    c_components = find_c_components(
        adjacency_matrix=adjacency_matrix_minus_x, node_set=node_names_minus_x, idx2node=idx2node_minus_x
    )
    if len(c_components) > 1:
        identifier = IDExpression()
        sum_over_set = node_names.difference(outcome_name.union(treatment_name))
        for component in c_components:
            expressions = __adjacency_matrix_identify_effect(
                treatment_name=node_names.difference(component),
                outcome_name=OrderedSet(list(component)),
                adjacency_matrix=adjacency_matrix,
                tsort_node_names=tsort_node_names,
                node_names=node_names,
            )
            for expression in expressions.get_val(return_type="prod"):
                identifier.add_product(expression)
        identifier.add_sum(sum_over_set)
        estimators.add_product(identifier)
        return estimators

    # Line 5 - The algorithms fails due to the presence of a hedge - the graph G, and a subgraph S that does not contain any X nodes.
    S = c_components[0]
    c_components_G = find_c_components(adjacency_matrix=adjacency_matrix, node_set=node_names, idx2node=idx2node)
    if len(c_components_G) == 1 and c_components_G[0] == node_names:
        return None

    # Line 6 - If there are no bidirected arcs from X to the other nodes in the current subproblem under consideration, then we can replace acting on X by conditioning, and thus solve the subproblem.
    if S in c_components_G:
        sum_over_set = S.difference(outcome_name)
        prev_nodes = []
        for node in tsort_node_names:
            if node in S:
                identifier = IDExpression()
                estimator = {}
                estimator["outcome_vars"] = OrderedSet([node])
                estimator["condition_vars"] = OrderedSet(prev_nodes)
                identifier.add_product(estimator)
                identifier.add_sum(sum_over_set)
                estimators.add_product(identifier)
            prev_nodes.append(node)
        return estimators

    # Line 7 - This is the most complicated case in the algorithm. Explain in the second last paragraph on Pg 41 of the link provided in the docstring above.
    for component in c_components_G:
        C = S.difference(component)
        if C.is_empty() is None:
            return __adjacency_matrix_identify_effect(
                treatment_name=treatment_name.intersection(component),
                outcome_name=outcome_name,
                adjacency_matrix=induced_graph(
                    node_set=component, adjacency_matrix=adjacency_matrix, node2idx=node2idx
                ),
                tsort_node_names=tsort_node_names,
                node_names=node_names,
            )


def __idx_node_mapping(node_names: OrderedSet):
    """
    Obtain the node name to index and index to node name mappings.

    :param node_names: Name of all nodes in the graph.
    :return: node to index and index to node mappings.
    """
    node2idx = {}
    idx2node = {}
    for i, node in enumerate(node_names.get_all()):
        node2idx[node] = i
        idx2node[i] = node
    return node2idx, idx2node



"""This module provides the APIs for attributing the change in the output value of a deterministic mechanism for a statistical unit.
"""

from abc import abstractmethod
from typing import List, Optional

import numpy as np
import pandas as pd
from sklearn.linear_model._base import LinearModel
from sklearn.utils.validation import check_is_fitted

from dowhy.gcm.fcms import PredictionModel
from dowhy.gcm.ml.regression import SklearnRegressionModel
from dowhy.gcm.shapley import ShapleyConfig, estimate_shapley_values


class LinearPredictionModel:
    @property
    @abstractmethod
    def coefficients(self) -> np.ndarray:
        pass


class SklearnLinearRegressionModel(SklearnRegressionModel, LinearPredictionModel):
    def __init__(self, sklearn_mdl: LinearModel) -> None:
        super(SklearnLinearRegressionModel, self).__init__(sklearn_mdl)

    @property
    def coefficients(self) -> np.ndarray:
        check_is_fitted(self.sklearn_model)
        return self.sklearn_model.coef_


def unit_change(
    background_df: pd.DataFrame,
    foreground_df: pd.DataFrame,
    input_column_names: List[str],
    background_mechanism: PredictionModel,
    foreground_mechanism: Optional[PredictionModel] = None,
    shapley_config: Optional[ShapleyConfig] = None,
) -> pd.DataFrame:
    """
    This function attributes the change in the output value of a deterministic mechanism for a statistical unit to each input and optionally for the mechanism if `foreground_mechanism` is provided.
    The technical method is described in the following research paper:
    Kailash Budhathoki, George Michailidis, Dominik Janzing. *Explaining the root causes of unit-level changes*. arXiv, 2022.

    :param background_df: The background dataset.
    :param foreground_df: The foreground dataset.
    :param input_column_names: The names of the input columns.
    :param background_mechanism: The background mechanism. If the mechanism does not change, then this mechanism is used for attribution.
    :param foreground_mechanism: The foreground mechanism. If provided, the method also attributes the output change to the change in the mechanism.
    :param shapley_config: The configuration for calculating Shapley values.
    :return: A dataframe containing the contributions of each input and optionally the mechanism to the change in the output values of the deterministic mechanism(s) for given inputs.
    """
    if foreground_mechanism:
        if isinstance(background_mechanism, LinearPredictionModel):
            return unit_change_linear(
                background_mechanism, background_df, foreground_mechanism, foreground_df, input_column_names
            )
        else:
            return unit_change_nonlinear(
                background_mechanism,
                background_df,
                foreground_mechanism,
                foreground_df,
                input_column_names,
                shapley_config,
            )

    if isinstance(background_mechanism, LinearPredictionModel):
        return unit_change_linear_input_only(background_mechanism, background_df, foreground_df, input_column_names)
    else:
        return unit_change_nonlinear_input_only(
            background_mechanism, background_df, foreground_df, input_column_names, shapley_config
        )


def unit_change_nonlinear(
    background_mechanism: PredictionModel,
    background_df: pd.DataFrame,
    foreground_mechanism: PredictionModel,
    foreground_df: pd.DataFrame,
    input_column_names: List[str],
    shapley_config: Optional[ShapleyConfig] = None,
) -> pd.DataFrame:
    """
    Calculates the contributions of mechanism and each input to the change in the output values of a non-linear deterministic mechanism.
    The technical method is described in the following research paper:
    Kailash Budhathoki, George Michailidis, Dominik Janzing. *Explaining the root causes of unit-level changes*. arXiv, 2022.

    :param background_mechanism: The background mechanism.
    :param background_df: The background data.
    :param foreground_mechanism: The foreground mechanism.
    :param foreground_df: The foreground data.
    :param input_column_names: The names of the input (features) columns in both dataframes.
    :param shapley_config: The configuration for calculating Shapley values.
    :return: A pandas dataframe with attributions to each cause for the change in each output row of provided dataframes.
    """
    _check_if_input_columns_exist(background_df, foreground_df, input_column_names)

    def payoff(player_indicator: List[int]) -> np.ndarray:
        """The last cell in the binary vector represents the player 'mechanism'."""
        input_arrays = []
        for i, is_player_active in enumerate(player_indicator[:-1]):
            selected_df = foreground_df if is_player_active else background_df
            input_arrays.append(selected_df[input_column_names[i]].to_numpy())
        mechanism = foreground_mechanism if player_indicator[-1] else background_mechanism
        return mechanism.predict(np.column_stack(input_arrays)).flatten()

    contributions = estimate_shapley_values(payoff, len(input_column_names) + 1, shapley_config)
    root_causes = input_column_names + ["f"]
    return pd.DataFrame(contributions, columns=root_causes)


def unit_change_linear(
    background_mechanism: LinearPredictionModel,
    background_df: pd.DataFrame,
    foreground_mechanism: LinearPredictionModel,
    foreground_df: pd.DataFrame,
    input_column_names: List[str],
) -> pd.DataFrame:
    """
    Calculates the contributions of mechanism and each input to the change in the output values of a linear deterministic mechanism.

    :param background_mechanism: The linear background mechanism.
    :param background_df: The background data.
    :param foreground_mechanism: The linear foreground mechanism.
    :param foreground_df: The foreground data.
    :param input_column_names: The names of the input columns in both dataframes.
    :return: A pandas dataframe with attributions to each cause for the change in each output row of provided dataframes.
    """
    _check_if_input_columns_exist(background_df, foreground_df, input_column_names)

    coeffs_total = background_mechanism.coefficients + foreground_mechanism.coefficients  # p x 1
    coeffs_diff = foreground_mechanism.coefficients - background_mechanism.coefficients  # p x 1

    input_total = foreground_df[input_column_names].to_numpy() + background_df[input_column_names].to_numpy()  # n x p
    input_diff = foreground_df[input_column_names].to_numpy() - background_df[input_column_names].to_numpy()  # n x p

    contribution_input = 0.5 * np.einsum("ij,ki->ki", coeffs_total.reshape(-1, 1), input_diff)
    contribution_mechanism = 0.5 * np.einsum("ij,ki->k", coeffs_diff.reshape(-1, 1), input_total)
    contribution_df = pd.DataFrame(contribution_input, columns=input_column_names)
    contribution_df["f"] = contribution_mechanism  # TODO: Handle the case where 'f' is an input column name
    return contribution_df


def unit_change_nonlinear_input_only(
    mechanism: PredictionModel,
    background_df: pd.DataFrame,
    foreground_df: pd.DataFrame,
    input_column_names: List[str],
    shapley_config: Optional[ShapleyConfig] = None,
) -> pd.DataFrame:
    """
    Calculates the contributions of each input to the change in the output values of a non-linear deterministic mechanism.
    The technical method is a modification of the attribution method described in the following research paper, without mechanism as a player:
    Kailash Budhathoki, George Michailidis, Dominik Janzing. *Explaining the root causes of unit-level changes*. arXiv, 2022.

    :param mechanism: The mechanism.
    :param background_df: The background data.
    :param foreground_df: The foreground data.
    :param input_column_names: The names of the input (features) columns in both dataframes.
    :param shapley_config: The configuration for calculating Shapley values.
    :return: A pandas dataframe with attributions to each cause for the change in each output row of provided dataframes.
    """
    _check_if_input_columns_exist(background_df, foreground_df, input_column_names)

    def payoff(player_indicator: List[int]) -> np.ndarray:
        input_arrays = []
        for i, is_player_active in enumerate(player_indicator):
            selected_df = foreground_df if is_player_active else background_df
            input_arrays.append(selected_df[input_column_names[i]].to_numpy())
        return mechanism.predict(np.column_stack(input_arrays)).flatten()

    contributions = estimate_shapley_values(payoff, len(input_column_names), shapley_config)
    return pd.DataFrame(contributions, columns=input_column_names)


def unit_change_linear_input_only(
    mechanism: LinearPredictionModel,
    background_df: pd.DataFrame,
    foreground_df: pd.DataFrame,
    input_column_names: List[str],
) -> pd.DataFrame:
    """
    Calculates the contributions of each input to the change in the output values of a linear deterministic mechanism.

    :param mechanism: The linear mechanism.
    :param background_df: The background data.
    :param foreground_df: The foreground data.
    :param input_column_names: The names of the input (features) columns in both dataframes.
    :return: A pandas dataframe with attributions to each cause for the change in each output row of provided dataframes.
    """
    _check_if_input_columns_exist(background_df, foreground_df, input_column_names)

    input_diff = foreground_df[input_column_names].to_numpy() - background_df[input_column_names].to_numpy()  # n x p
    contribution_input = np.einsum("ij,ki->ki", mechanism.coefficients.reshape(-1, 1), input_diff)
    return pd.DataFrame(contribution_input, columns=input_column_names)


def _check_if_input_columns_exist(
    background_df: pd.DataFrame, foreground_df: pd.DataFrame, input_column_names: List[str]
) -> None:
    if not len(set(background_df.columns).intersection(input_column_names)) == len(input_column_names) or not len(
        set(foreground_df.columns).intersection(input_column_names)
    ) == len(input_column_names):
        raise ValueError("Input column names not found in either the background or the foreground data.")



import logging
import random
from typing import List, Optional, Union

import numpy as np
import pandas as pd
from joblib import Parallel, delayed
from sklearn.utils import resample
from tqdm.auto import tqdm

from dowhy.causal_estimator import CausalEstimate, CausalEstimator
from dowhy.causal_estimators.econml import Econml
from dowhy.causal_identifier.identified_estimand import IdentifiedEstimand
from dowhy.causal_refuter import CausalRefutation, CausalRefuter, choose_variables, test_significance

logger = logging.getLogger(__name__)


class BootstrapRefuter(CausalRefuter):
    """
    Refute an estimate by running it on a random sample of the data containing measurement error in the
    confounders. This allows us to find the ability of the estimator to find the effect of the
    treatment on the outcome.

    It supports additional parameters that can be specified in the refute_estimate() method.

    :param num_simulations: The number of simulations to be run, ``CausalRefuter.DEFAULT_NUM_SIMULATIONS`` by default
    :type num_simulations: int, optional

    :param sample_size: The size of each bootstrap sample and is the size of the original data by default
    :type sample_size: int, optional

    :param required_variables: The list of variables to be used as the input for ``y~f(W)``
      This is ``True`` by default, which in turn selects all variables leaving the treatment and the outcome
    :type required_variables: int, list, bool, optional

    1. An integer argument refers to how many variables will be used for estimating the value of the outcome
    2. A list explicitly refers to which variables will be used to estimate the outcome
       Furthermore, it gives the ability to explictly select or deselect the covariates present in the estimation of the
       outcome. This is done by either adding or explicitly removing variables from the list as shown below:

    .. note::
            * We need to pass required_variables = ``[W0,W1]`` if we want ``W0`` and ``W1``.
            * We need to pass required_variables = ``[-W0,-W1]`` if we want all variables excluding ``W0`` and ``W1``.

    3. If the value is True, we wish to include all variables to estimate the value of the outcome.

    .. warning:: A ``False`` value is ``INVALID`` and will result in an ``error``.

    :param noise: The standard deviation of the noise to be added to the data and is ``BootstrapRefuter.DEFAULT_STD_DEV`` by default
    :type noise: float, optional

    :param probability_of_change: It specifies the probability with which we change the data for a boolean or categorical variable
      It is ``noise`` by default, only if the value of ``noise`` is less than 1.
    :type probability_of_change: float, optional

    :param random_state: The seed value to be added if we wish to repeat the same random behavior. For this purpose, we repeat the same seed in the psuedo-random generator.
    :type random_state: int, RandomState, optional
    """

    DEFAULT_STD_DEV = 0.1
    DEFAULT_SUCCESS_PROBABILITY = 0.5
    DEFAULT_NUMBER_OF_TRIALS = 1

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._num_simulations = kwargs.pop("num_simulations", CausalRefuter.DEFAULT_NUM_SIMULATIONS)
        self._sample_size = kwargs.pop("sample_size", len(self._data))
        self._required_variables = kwargs.pop("required_variables", True)
        self._noise = kwargs.pop("noise", BootstrapRefuter.DEFAULT_STD_DEV)
        self._probability_of_change = kwargs.pop("probability_of_change", None)
        self._random_state = kwargs.pop("random_state", None)

        self.logger = logging.getLogger(__name__)

    def refute_estimate(self, show_progress_bar: bool = False, *args, **kwargs):
        refute = refute_bootstrap(
            data=self._data,
            target_estimand=self._target_estimand,
            estimate=self._estimate,
            num_simulations=self._num_simulations,
            random_state=self._random_state,
            sample_size=self._sample_size,
            required_variables=self._required_variables,
            noise=self._noise,
            probability_of_change=self._probability_of_change,
            show_progress_bar=show_progress_bar,
            n_jobs=self._n_jobs,
            verbose=self._verbose,
        )
        refute.add_refuter(self)
        return refute


def _refute_once(
    data: pd.DataFrame,
    target_estimand: IdentifiedEstimand,
    estimate: CausalEstimate,
    chosen_variables: Optional[List] = None,
    random_state: Optional[Union[int, np.random.RandomState]] = None,
    sample_size: Optional[int] = None,
    noise: float = 0.1,
    probability_of_change: Optional[float] = None,
):
    if random_state is None:
        new_data = resample(data, n_samples=sample_size)
    else:
        new_data = resample(data, n_samples=sample_size, random_state=random_state)

    if chosen_variables is not None:
        for variable in chosen_variables:

            if ("float" or "int") in new_data[variable].dtype.name:
                scaling_factor = new_data[variable].std()
                new_data[variable] += np.random.normal(loc=0.0, scale=noise * scaling_factor, size=sample_size)

            elif "bool" in new_data[variable].dtype.name:
                probs = np.random.uniform(0, 1, sample_size)
                new_data[variable] = np.where(
                    probs < probability_of_change, np.logical_not(new_data[variable]), new_data[variable]
                )

            elif "category" in new_data[variable].dtype.name:
                categories = new_data[variable].unique()
                # Find the set difference for each row
                changed_data = new_data[variable].apply(lambda row: list(set(categories) - set([row])))
                # Choose one out of the remaining
                changed_data = changed_data.apply(lambda row: random.choice(row))
                new_data[variable] = np.where(probs < probability_of_change, changed_data)
                new_data[variable].astype("category")

    new_estimator = estimate.estimator.get_new_estimator_object(target_estimand)
    new_estimator.fit(
        new_data,
        target_estimand.treatment_variable,
        target_estimand.outcome_variable,
        estimate.estimator._effect_modifier_names,
        **new_estimator._econml_fit_params if isinstance(new_estimator, Econml) else {},
    )
    new_effect = new_estimator.estimate_effect(
        control_value=estimate.control_value,
        treatment_value=estimate.treatment_value,
        target_units=estimate.estimator._target_units,
    )
    return new_effect.value


def refute_bootstrap(
    data: pd.DataFrame,
    target_estimand: IdentifiedEstimand,
    estimate: CausalEstimate,
    num_simulations: int = 100,
    random_state: Optional[Union[int, np.random.RandomState]] = None,
    sample_size: Optional[int] = None,
    required_variables: bool = True,
    noise: float = 0.1,
    probability_of_change: Optional[float] = None,
    show_progress_bar: bool = False,
    n_jobs: int = 1,
    verbose: int = 0,
    **_,
) -> CausalRefutation:
    """Refute an estimate by running it on a random sample of the data containing measurement error in the
    confounders. This allows us to find the ability of the estimator to find the effect of the
    treatment on the outcome.

    :param data: pd.DataFrame: Data to run the refutation
    :param target_estimand: IdentifiedEstimand: Identified estimand to run the refutation
    :param estimate: CausalEstimate: Estimate to run the refutation
    :param num_simulations: The number of simulations to be run, ``CausalRefuter.DEFAULT_NUM_SIMULATIONS`` by default
    :param random_state: The seed value to be added if we wish to repeat the same random behavior. For this purpose, we repeat the same seed in the psuedo-random generator.
    :param sample_size: The size of each bootstrap sample and is the size of the original data by default
    :param required_variables: The list of variables to be used as the input for ``y~f(W)``
      This is ``True`` by default, which in turn selects all variables leaving the treatment and the outcome
    1. An integer argument refers to how many variables will be used for estimating the value of the outcome
    2. A list explicitly refers to which variables will be used to estimate the outcome
       Furthermore, it gives the ability to explictly select or deselect the covariates present in the estimation of the
       outcome. This is done by either adding or explicitly removing variables from the list as shown below:
    .. note::
            * We need to pass required_variables = ``[W0,W1]`` if we want ``W0`` and ``W1``.
            * We need to pass required_variables = ``[-W0,-W1]`` if we want all variables excluding ``W0`` and ``W1``.
    3. If the value is True, we wish to include all variables to estimate the value of the outcome.
    .. warning:: A ``False`` value is ``INVALID`` and will result in an ``error``.
    :param noise: The standard deviation of the noise to be added to the data and is ``BootstrapRefuter.DEFAULT_STD_DEV`` by default
    :param probability_of_change: It specifies the probability with which we change the data for a boolean or categorical variable
      It is ``noise`` by default, only if the value of ``noise`` is less than 1.
    :param n_jobs: The maximum number of concurrently running jobs. If -1 all CPUs are used. If 1 is given, no parallel computing code is used at all (this is the default).
    :param verbose: The verbosity level: if non zero, progress messages are printed. Above 50, the output is sent to stdout. The frequency of the messages increases with the verbosity level. If it more than 10, all iterations are reported. The default is 0.
    """
    if sample_size is None:
        sample_size = len(data)

    chosen_variables = choose_variables(
        required_variables,
        target_estimand.get_backdoor_variables()
        + target_estimand.instrumental_variables
        + estimate.estimator._effect_modifier_names,
    )

    if chosen_variables is None:
        logger.info("INFO: There are no chosen variables")
    else:
        logger.info("INFO: The chosen variables are: " + ",".join(chosen_variables))

    if probability_of_change is None and noise > 1:
        logger.error("Error in using noise:{} for Binary Flip. The value is greater than 1".format(noise))
        raise ValueError("The value for Binary Flip cannot be greater than 1")
    elif probability_of_change is None and noise <= 1:
        probability_of_change = noise
    elif probability_of_change > 1:
        logger.error(
            "The probability of flip is: {}, However, this value cannot be greater than 1".format(probability_of_change)
        )
        raise ValueError("Probability of Flip cannot be greater than 1")

    if sample_size > len(data):
        logger.warning("The sample size is larger than the population size")

    logger.info("Refutation over {} simulated datasets of size {} each".format(num_simulations, sample_size))

    sample_estimates = Parallel(n_jobs=n_jobs, verbose=verbose)(
        delayed(_refute_once)(
            data, target_estimand, estimate, chosen_variables, random_state, sample_size, noise, probability_of_change
        )
        for _ in tqdm(
            range(num_simulations),
            colour=CausalRefuter.PROGRESS_BAR_COLOR,
            disable=not show_progress_bar,
            desc="Refuting Estimates: ",
        )
    )
    sample_estimates = np.array(sample_estimates)

    refute = CausalRefutation(
        estimate.value, np.mean(sample_estimates), refutation_type="Refute: Bootstrap Sample Dataset"
    )

    # We want to see if the estimate falls in the same distribution as the one generated by the refuter
    # Ideally that should be the case as running bootstrap should not have a significant effect on the ability
    # of the treatment to affect the outcome
    refute.add_significance_test_results(test_significance(estimate, sample_estimates))

    return refute



"""Contains a method to reject the causal graph and validate causal mechanisms such as post non-linear models.

Classes and functions in this module should be considered experimental, meaning there might be breaking API changes in
the future.
"""
from enum import Enum, auto
from typing import Any, Callable, Dict, Optional, Set, Tuple, Union

import networkx as nx
import numpy as np
import pandas as pd
from statsmodels.stats.multitest import multipletests

from dowhy.gcm.cms import InvertibleStructuralCausalModel
from dowhy.gcm.graph import DirectedGraph, get_ordered_predecessors, is_root_node, validate_causal_graph
from dowhy.gcm.independence_test import kernel_based


class RejectionResult(Enum):
    REJECTED = auto()
    NOT_REJECTED = (auto(),)


def refute_causal_structure(
    causal_graph: DirectedGraph,
    data: pd.DataFrame,
    independence_test: Callable[[np.ndarray, np.ndarray], float] = kernel_based,
    conditional_independence_test: Callable[[np.ndarray, np.ndarray, np.ndarray], float] = kernel_based,
    significance_level: float = 0.05,
    fdr_control_method: Optional[str] = "fdr_bh",
) -> Tuple[RejectionResult, Dict[str, Dict[str, Dict[str, Union[bool, float, Dict[str, Union[bool, float]]]]]]]:
    """Validates the assumptions in a causal graph against data. To this end, at each node, we test if the node is dependent on each of its parents, and test the local Markov condition.
    Note that valid local Markov conditions also imply a valid global Markov condition.

    :param causal_graph: A directed acyclic graph (DAG).
    :param data: Observations of variables in the DAG.
    :param independence_test: Independence test to use for checking edge dependencies.
    :param conditional_independence_test: Conditional independence test to use for checking local Markov condition.
    :param significance_level: Significance level for (conditional) independence tests.
    :param fdr_control_method: Method for false discovery rate (FDR) control. For various options, please refer to `this page <https://www.statsmodels.org/dev/generated/statsmodels.stats.multitest.multipletests.html>`_.
    :return: Outcome of the validation process. The first element of the tuple indicates whether the graph is valid w.r.t. given data, and the second element gives the summary of tests at each node. An example for X->Y->Z:

    .. code-block:: python

        [True, {'X': {'local_markov_test': {}, 'edge_dependence_test': {}},
                'Y': {'local_markov_test': {}, 'edge_dependence_test': {'X': {'p_value': 0.5, 'fdr_adjusted_p_value': 0.5, 'success': True}}},
                'Z': {'local_markov_test': {'p_value': 0.0, 'fdr_adjusted_p_value': 0.5, 'success': False},
                      'edge_dependence_test': {'Y': {'p_value': 0.5, 'fdr_adjusted_p_value': 0.5, 'success': True}}}}]
    """
    is_dag_valid = True
    validation_summary = dict()
    all_p_values = []

    for node in causal_graph.nodes:
        parents = get_ordered_predecessors(causal_graph, node)
        non_descendants = _get_non_descendants(causal_graph, node, exclude_parents=True)

        lmc_test_result = dict()
        if parents and non_descendants:
            # test local Markov condition, null hypothesis: conditional independence
            lmc_p_value = conditional_independence_test(
                data[node].values, data[non_descendants].values, data[parents].values
            )
            lmc_test_result = dict(p_value=lmc_p_value)
            all_p_values.append(lmc_p_value)

        # test edge dependence, null hypothesis: independence
        edge_dependence_result = dict()
        for parent in parents:
            edge_p_value = independence_test(data[parent].values, data[node].values)
            edge_dependence_result[parent] = dict(p_value=edge_p_value)
            all_p_values.append(edge_p_value)

        validation_summary[node] = dict(local_markov_test=lmc_test_result, edge_dependence_test=edge_dependence_result)

    if fdr_control_method is None:
        successes = np.array(all_p_values) <= significance_level
        adjusted_p_values = np.empty(len(successes))
        adjusted_p_values[:] = np.nan
    else:
        multipletests_result = multipletests(all_p_values, significance_level, method=fdr_control_method)
        successes = multipletests_result[0]
        adjusted_p_values = multipletests_result[1]

    # The order of the p-values added to the list is deterministic.
    index = 0
    for node in causal_graph.nodes:
        if "p_value" in validation_summary[node]["local_markov_test"]:
            validation_summary[node]["local_markov_test"]["fdr_adjusted_p_value"] = adjusted_p_values[index]
            validation_summary[node]["local_markov_test"]["success"] = not successes[index]
            is_dag_valid &= not successes[index]
            index += 1

        for parent in get_ordered_predecessors(causal_graph, node):
            validation_summary[node]["edge_dependence_test"][parent]["fdr_adjusted_p_value"] = adjusted_p_values[index]
            validation_summary[node]["edge_dependence_test"][parent]["success"] = successes[index]
            is_dag_valid &= successes[index]
            index += 1
    if is_dag_valid:
        return RejectionResult.NOT_REJECTED, validation_summary
    else:
        return RejectionResult.REJECTED, validation_summary


def refute_invertible_model(
    causal_model: InvertibleStructuralCausalModel,
    data: pd.DataFrame,
    independence_test: Callable[[np.ndarray, np.ndarray], float] = kernel_based,
    significance_level: float = 0.05,
    fdr_control_method: Optional[str] = None,
) -> RejectionResult:
    """Validate the assumption that the structural causal models can be represented by a
    :py:class:`InvertibleFunctionalCausalModel <dowhy.gcm.graph.InvertibleFunctionalCausalModel>` (e.g. the causal mechanisms are
    :py:class:`AdditiveNoiseModels <dowhy.gcm.AdditiveNoiseModel>` and/or :py:class:`PostNonlinearModels <dowhy.gcm.PostNonlinearModel>`).
    For this, it is checked if the residual of a causal mechanism is independent of the mechanism's input
    (i.e. we assume causal sufficiency here). For instance, :py:class:`PostNonlinearModels <dowhy.gcm.PostNonlinearModel>` represent
        Y = f(g(X) + N),
    where f is invertible (g does not need to be), X are the parents of Y and N is (assumed to be) independent noise.
    The latter point is important here. For given data, we can then reconstruct N and perform an independence test between X and N.

    Note that this method only validates the causal mechanisms and not the graph structure.

    For the case of post non-linear models, see the following paper for more details:
        Zhang, K., and A. Hyvärinen.
        On the Identifiability of the Post-Nonlinear Causal Model.
        25th Conference on Uncertainty in Artificial Intelligence (UAI 2009). AUAI Press, 2009.

    :param causal_model: A fitted invertible structural causal model.
    :param data: Observations of variables in the DAG.
    :param independence_test: Independence test to use for checking if residual and input are dependent.
    :param significance_level: Significance level for deciding whether input and residual is dependent.
    :param fdr_control_method: Method for false discovery rate (FDR) control. For various options, please refer to `this page <https://www.statsmodels.org/dev/generated/statsmodels.stats.multitest.multipletests.html>`_.
    :return: The outcome of the validation. The causal model can not be rejected if all causal mechanisms are consistent with
             the invertible model assumption.
    """
    validate_causal_graph(causal_model.graph)

    all_p_values = []

    for node in causal_model.graph.nodes:
        if is_root_node(causal_model.graph, node):
            continue
        else:
            parents_samples = data[get_ordered_predecessors(causal_model.graph, node)].to_numpy()
            residuals = causal_model.causal_mechanism(node).estimate_noise(data[node].to_numpy(), parents_samples)

            all_p_values.append(independence_test(parents_samples, residuals))

    if fdr_control_method is None or len(all_p_values) == 0:
        return (
            RejectionResult.NOT_REJECTED
            if np.all(np.array(all_p_values) > significance_level)
            else RejectionResult.REJECTED
        )
    else:
        return (
            RejectionResult.REJECTED
            if np.any(multipletests(all_p_values, significance_level, method=fdr_control_method)[0])
            else RejectionResult.NOT_REJECTED
        )


def _get_non_descendants(causal_graph: DirectedGraph, node: Any, exclude_parents: bool = False) -> Set[Any]:
    nodes_to_exclude = nx.descendants(causal_graph, node).union({node})
    if exclude_parents:
        nodes_to_exclude = nodes_to_exclude.union(causal_graph.predecessors(node))
    return set(causal_graph.nodes).difference(nodes_to_exclude)



TEST_EFFICIENT_BD_SOLUTIONS = {
    # For all examples from these papers we use X for the treatment variable
    # instead of A.
    # Figure 6 from Smucler, Sapienza and Rotnitzky (2021), Biometrika
    "ssr21_fig6_example_graph": dict(
        graph_str="""graph[directed 1 node[id "X" label "X"]
                        node[id "F" label "F"]
                        node[id "T" label "T"]
                        node[id "U" label "U"]
                        node[id "Y" label "Y"]
                        edge[source "T" target "X"]
                        edge[source "X" target "Y"]
                        edge[source "U" target "Y"]
                        edge[source "U" target "F"]
                        ]
                        """,
        observed_node_names=["X", "Y", "T", "F"],
        conditional_node_names=["T"],
        efficient_adjustment=None,
        efficient_minimal_adjustment={"T"},
        efficient_mincost_adjustment={"T"},
        costs=None,
    ),
    # Figure 5 from Smucler, Sapienza and Rotnitzky (2021), Biometrika
    "ssr21_fig5_example_graph": dict(
        graph_str="""graph[directed 1 node[id "X" label "X"]
                        node[id "Y" label "Y"]
                        node[id "U" label "U"]
                        node[id "Z1" label "Z1"]
                        node[id "Z2" label "Z2"]
                        edge[source "X" target "Y"]
                        edge[source "Z1" target "X"]
                        edge[source "Z1" target "Z2"]
                        edge[source "U" target "Z2"]
                        edge[source "U" target "Y"]]
                        """,
        observed_node_names=["X", "Y", "Z1", "Z2"],
        conditional_node_names=[],
        efficient_adjustment=None,
        efficient_minimal_adjustment=set(),
        efficient_mincost_adjustment=set(),
        costs=None,
    ),
    # Figure 4 from Smucler, Sapienza and Rotnitzky (2021), Biometrika
    "ssr21_fig4_example_graph": dict(
        graph_str="""graph[directed 1 node[id "X" label "X"]
                        node[id "Y" label "Y"]
                        node[id "T" label "T"]
                        node[id "W1" label "W1"]
                        node[id "W2" label "W2"]
                        node[id "W3" label "W3"]
                        node[id "W4" label "W4"]
                        edge[source "X" target "Y"]
                        edge[source "T" target "W1"]
                        edge[source "T" target "Y"]
                        edge[source "W1" target "X"]
                        edge[source "W2" target "W1"]
                        edge[source "W2" target "Y"]
                        edge[source "W3" target "W1"]
                        edge[source "W3" target "Y"]
                        edge[source "W4" target "Y"]
                        ]
                        """,
        observed_node_names=["A", "Y", "T", "W1", "W2", "W3", "W4"],
        conditional_node_names=["T"],
        efficient_adjustment={"T", "W2", "W3", "W4"},
        efficient_minimal_adjustment={"T", "W2", "W3"},
        efficient_mincost_adjustment={"T", "W1"},
        costs=None,
    ),
    # Figure 3 from Smucler, Sapienza and Rotnitzky (2021), Biometrika
    "ssr21_fig3_example_graph": dict(
        graph_str="""graph[directed 1 node[id "X" label "X"]
                        node[id "F" label "F"]
                        node[id "T" label "T"]
                        node[id "U" label "U"]
                        node[id "Y" label "Y"]
                        node[id "M" label "M"]
                        edge[source "T" target "X"]
                        edge[source "T" target "F"]
                        edge[source "F" target "X"]
                        edge[source "U" target "F"]
                        edge[source "U" target "Y"]
                        edge[source "X" target "M"]
                        edge[source "M" target "Y"]
                        ]
                        """,
        observed_node_names=["X", "Y", "T", "F", "M"],
        conditional_node_names=["T"],
        efficient_adjustment={"T", "F"},
        efficient_minimal_adjustment={"T", "F"},
        efficient_mincost_adjustment={"T", "F"},
        costs=None,
    ),
    # Figure 2 from Smucler and Rotnitzky (2022), Journal of Causal Inference
    # L replaces X as the conditional variable
    "sr22_fig2_example_graph": dict(
        graph_str="""graph[directed 1 node[id "L" label "L"]
                        node[id "X" label "X"]
                        node[id "K" label "K"]
                        node[id "B" label "B"]
                        node[id "Q" label "Q"]
                        node[id "R" label "R"]
                        node[id "T" label "T"]
                        node[id "M" label "M"]
                        node[id "Y" label "Y"]
                        node[id "U" label "U"]
                        node[id "F" label "F"]
                        edge[source "L" target "X"]
                        edge[source "X" target "M"]
                        edge[source "K" target "X"]
                        edge[source "B" target "K"]
                        edge[source "B" target "R"]
                        edge[source "Q" target "K"]
                        edge[source "Q" target "T"]
                        edge[source "R" target "Y"]
                        edge[source "T" target "Y"]
                        edge[source "M" target "Y"]
                        edge[source "U" target "Y"]
                        edge[source "U" target "F"]
                        ]
                        """,
        observed_node_names=["L", "X", "B", "K", "Q", "R", "M", "T", "Y", "F"],
        conditional_node_names=["L"],
        efficient_adjustment=None,
        efficient_minimal_adjustment={"L", "T", "R"},
        efficient_mincost_adjustment={"L", "T", "R"},
        costs=[
            ("L", {"cost": 1}),
            ("B", {"cost": 2}),
            ("K", {"cost": 4}),
            ("Q", {"cost": 1}),
            ("R", {"cost": 1}),
            ("T", {"cost": 1}),
        ],
    ),
    # Figure 2 from Smucler and Rotnitzky (2022), Journal of Causal Inference
    # L replaces X as the conditional variable. Uses different costs
    "sr22_fig2_diffcosts_example_graph": dict(
        graph_str="""graph[directed 1 node[id "L" label "L"]
                        node[id "X" label "X"]
                        node[id "K" label "K"]
                        node[id "B" label "B"]
                        node[id "Q" label "Q"]
                        node[id "R" label "R"]
                        node[id "T" label "T"]
                        node[id "M" label "M"]
                        node[id "Y" label "Y"]
                        node[id "U" label "U"]
                        node[id "F" label "F"]
                        edge[source "L" target "X"]
                        edge[source "X" target "M"]
                        edge[source "K" target "X"]
                        edge[source "B" target "K"]
                        edge[source "B" target "R"]
                        edge[source "Q" target "K"]
                        edge[source "Q" target "T"]
                        edge[source "R" target "Y"]
                        edge[source "T" target "Y"]
                        edge[source "M" target "Y"]
                        edge[source "U" target "Y"]
                        edge[source "U" target "F"]
                        ]
                        """,
        observed_node_names=["L", "X", "B", "K", "Q", "R", "M", "T", "Y", "F"],
        conditional_node_names=["L"],
        efficient_adjustment=None,
        efficient_minimal_adjustment={"L", "T", "R"},
        efficient_mincost_adjustment={"L", "T", "B"},
        costs=[
            ("L", {"cost": 1}),
            ("B", {"cost": 1}),
            ("K", {"cost": 4}),
            ("Q", {"cost": 1}),
            ("R", {"cost": 2}),
            ("T", {"cost": 1}),
        ],
    ),
    # Figure 3 from Smucler and Rotnitzky (2022), Journal of Causal Inference
    "sr22_fig3_example_graph": dict(
        graph_str="""graph[directed 1 node[id "X" label "X"]
                        node[id "B" label "B"]
                        node[id "Q" label "Q"]
                        node[id "R" label "R"]
                        node[id "T" label "T"]
                        node[id "Y" label "Y"]
                        edge[source "B" target "X"]
                        edge[source "Q" target "X"]
                        edge[source "B" target "T"]
                        edge[source "Q" target "R"]
                        edge[source "T" target "Y"]
                        edge[source "R" target "Y"]
                        edge[source "X" target "Y"]
                        ]
                        """,
        observed_node_names=["X", "B", "Q", "R", "T", "Y"],
        conditional_node_names=[],
        efficient_adjustment={"R", "T"},
        efficient_minimal_adjustment={"R", "T"},
        efficient_mincost_adjustment={"B", "Q"},
        costs=[
            ("B", {"cost": 1}),
            ("Q", {"cost": 1}),
            ("R", {"cost": 2}),
            ("T", {"cost": 2}),
        ],
    ),
    # A graph where optimal, optimal minimal and optimal min cost are different
    "alldiff_example_graph": dict(
        graph_str="""graph[directed 1 node[id "X" label "X"]
                        node[id "Y" label "Y"]
                        node[id "W1" label "W1"]
                        node[id "W2" label "W2"]
                        node[id "W3" label "W3"]
                        node[id "W4" label "W4"]
                        node[id "W5" label "W5"]
                        node[id "W6" label "W6"]
                        node[id "W7" label "W7"]
                        node[id "W8" label "W8"]
                        node[id "W9" label "W9"]
                        node[id "K" label "K"]
                        node[id "O" label "O"]
                        edge[source "K" target "X"]
                        edge[source "X" target "Y"]
                        edge[source "W1" target "K"]
                        edge[source "W2" target "K"]
                        edge[source "W3" target "K"]
                        edge[source "W1" target "W4"]
                        edge[source "W2" target "W5"]
                        edge[source "W3" target "W6"]
                        edge[source "W4" target "W7"]
                        edge[source "W5" target "W8"]
                        edge[source "W6" target "W9"]
                        edge[source "W7" target "Y"]
                        edge[source "W8" target "Y"]
                        edge[source "W9" target "Y"]
                        edge[source "O" target "Y"]
                        ]
                        """,
        observed_node_names=[
            "X",
            "Y",
            "K",
            "O",
            "W1",
            "W2",
            "W3",
            "W4",
            "W5",
            "W6",
            "W7",
            "W8",
            "W9",
        ],
        conditional_node_names=[],
        efficient_adjustment={"O", "W7", "W8", "W9"},
        efficient_minimal_adjustment={"W7", "W8", "W9"},
        efficient_mincost_adjustment={"W7", "W8", "W6"},
        costs=[
            ("W1", {"cost": 2}),
            ("W2", {"cost": 1}),
            ("W3", {"cost": 1}),
            ("W4", {"cost": 1}),
            ("W5", {"cost": 2}),
            ("W6", {"cost": 1}),
            ("W7", {"cost": 1}),
            ("W8", {"cost": 1}),
            ("W9", {"cost": 2}),
            ("K", {"cost": 11}),
            ("O", {"cost": 1}),
        ],
    ),
    # The graph from Shrier and Platt (2008)
    "shrier_platt_2008": dict(
        graph_str="""graph[directed 1 node[id "coach" label "coach"]
                        node[id "team motivation" label "team motivation"]
                        node[id "fitness" label "fitness"]
                        node[id "pre-game prop" label "pre-game prop"]
                        node[id "intra-game prop" label "intra-game prop"]                       
                        node[id "neuromusc fatigue" label "neuromusc fatigue"]
                        node[id "X" label "X"]
                        node[id "previous injury" label "previous injury"]
                        node[id "contact sport" label "contact sport"]
                        node[id "genetics" label "genetics"]
                        node[id "Y" label "Y"]
                        node[id "tissue disorder" label "tissue disorder"]
                        node[id "tissue weakness" label "tissue weakness"]
                        edge[source "coach" target "team motivation"]
                        edge[source "coach" target "fitness"]
                        edge[source "fitness" target "pre-game prop"]
                        edge[source "fitness" target "neuromusc fatigue"]
                        edge[source "team motivation" target "X"]
                        edge[source "team motivation" target "previous injury"]
                        edge[source "pre-game prop" target "X"]
                        edge[source "X" target "intra-game prop"]
                        edge[source "contact sport" target "previous injury"]
                        edge[source "contact sport" target "intra-game prop"]
                        edge[source "intra-game prop" target "Y"]
                        edge[source "genetics" target "fitness"]
                        edge[source "genetics" target "neuromusc fatigue"]
                        edge[source "genetics" target "tissue disorder"]
                        edge[source "tissue disorder" target "neuromusc fatigue"]
                        edge[source "tissue disorder" target "tissue weakness"]
                        edge[source "neuromusc fatigue" target "intra-game prop"]
                        edge[source "neuromusc fatigue" target "Y"]
                        edge[source "tissue weakness" target "Y"]
                        ]
""",
        observed_node_names=[
            "team motivation",
            "previous injury",
            "X",
            "coach",
            "fitness",
            "contact sport",
            "neuromusc fatigue",
            "tissue disorder",
            "Y",
        ],
        conditional_node_names=["previous injury", "team motivation"],
        efficient_adjustment={
            "team motivation",
            "previous injury",
            "contact sport",
            "tissue disorder",
            "neuromusc fatigue",
        },
        efficient_minimal_adjustment={
            "team motivation",
            "previous injury",
            "tissue disorder",
            "neuromusc fatigue",
        },
        efficient_mincost_adjustment={"team motivation", "previous injury", "fitness"},
        costs=None,
    ),
    # A graph for which the algorithm was producing wrong result due to a bug reported by Sara Taheri
    "taheri_graph1": dict(
        graph_str="""graph[directed 1 node[id "Z1" label "Z1"]
                        node[id "Z2" label "Z2"]
                        node[id "Z3" label "Z3"]
                        node[id "M1" label "M1"]
                        node[id "M2" label "M2"]                       
                        node[id "X" label "X"]
                        node[id "Y" label "Y"]
                        edge[source "Z1" target "Z2"]
                        edge[source "Z2" target "Z3"]
                        edge[source "Z3" target "Y"]
                        edge[source "Z1" target "X"]
                        edge[source "X" target "M1"]
                        edge[source "M1" target "M2"]
                        edge[source "M2" target "Y"]
                        ]
""",
        observed_node_names=["Z1", "Z2", "Z3", "X", "Y", "M1", "M2"],
        conditional_node_names=[],
        efficient_adjustment={"Z3"},
        efficient_minimal_adjustment={"Z3"},
        efficient_mincost_adjustment={"Z3"},
        costs=None,
    ),
    # Another graph for which the algorithm was producing wrong result due to a bug reported by Sara Taheri
    "taheri_graph2": dict(
        graph_str="""graph[directed 1 node[id "Z1" label "Z1"]
                        node[id "Z2" label "Z2"]
                        node[id "Z3" label "Z3"]
                        node[id "Z4" label "Z4"]
                        node[id "Z5" label "Z5"]
                        node[id "M1" label "M1"]
                        node[id "M2" label "M2"]                       
                        node[id "M3" label "M3"]                       
                        node[id "X" label "X"]
                        node[id "Y" label "Y"]
                        node[id "U1" label "U1"]
                        node[id "U2" label "U2"]
                        edge[source "Z1" target "Z2"]
                        edge[source "Z2" target "Z3"]
                        edge[source "Z3" target "Z4"]
                        edge[source "Z4" target "Z5"]
                        edge[source "Z5" target "Y"]
                        edge[source "Z1" target "X"]
                        edge[source "X" target "M1"]
                        edge[source "M1" target "M2"]
                        edge[source "M2" target "M3"]
                        edge[source "M3" target "Y"]
                        edge[source "U1" target "X"]
                        edge[source "U1" target "Z1"]
                        edge[source "U2" target "Z2"]
                        edge[source "U2" target "M1"]
                        ]
""",
        observed_node_names=["Z1", "Z2", "Z3", "Z4", "Z5", "X", "Y", "M1", "M2", "M3"],
        conditional_node_names=[],
        efficient_adjustment={"Z1", "Z2", "Z5"},
        efficient_minimal_adjustment={"Z1"},
        efficient_mincost_adjustment={"Z1"},
        costs=None,
    ),
}



"""The gcm sub-package provides features built on top of graphical causal model (GCM) based inference. The status of
this addition and its API is considered experimental, meaning there might be breaking changes to its API in the
future.
"""

from . import auto, config, divergence, ml, shapley, stats, uncertainty, util
from .anomaly import anomaly_scores, attribute_anomalies
from .anomaly_scorers import (
    InverseDensityScorer,
    ITAnomalyScorer,
    MeanDeviationScorer,
    MedianCDFQuantileScorer,
    MedianDeviationScorer,
    RescaledMedianCDFQuantileScorer,
)
from .cms import FunctionalCausalModel, InvertibleStructuralCausalModel, ProbabilisticCausalModel, StructuralCausalModel
from .confidence_intervals import confidence_intervals
from .confidence_intervals_cms import bootstrap_sampling, fit_and_compute
from .density_estimators import GaussianMixtureDensityEstimator, KernelDensityEstimator1D
from .distribution_change import distribution_change, distribution_change_of_graphs
from .fcms import AdditiveNoiseModel, ClassificationModel, ClassifierFCM, PostNonlinearModel, PredictionModel
from .feature import feature_relevance_distribution, feature_relevance_sample, parent_relevance
from .fitting_sampling import draw_samples, fit
from .graph import ConditionalStochasticModel, DirectedGraph, FunctionalCausalModel, StochasticModel, is_root_node
from .independence_test import (
    approx_kernel_based,
    generalised_cov_based,
    independence_test,
    kernel_based,
    regression_based,
)
from .influence import arrow_strength, intrinsic_causal_influence
from .stochastic_models import BayesianGaussianMixtureDistribution, EmpiricalDistribution, ScipyDistribution
from .unit_change import unit_change
from .validation import RejectionResult, refute_causal_structure, refute_invertible_model
from .whatif import average_causal_effect, counterfactual_samples, interventional_samples



from dowhy.causal_identifier.auto_identifier import (
    AutoIdentifier,
    BackdoorAdjustment,
    EstimandType,
    identify_effect_auto,
)
from dowhy.causal_identifier.id_identifier import IDIdentifier, identify_effect_id
from dowhy.causal_identifier.identified_estimand import IdentifiedEstimand
from dowhy.causal_identifier.identify_effect import identify_effect

__all__ = [
    "AutoIdentifier",
    "identify_effect_auto",
    "identify_effect_id",
    "BackdoorAdjustment",
    "EstimandType",
    "IdentifiedEstimand",
    "IDIdentifier",
    "identify_effect",
]



import numpy as np
import pytest
from flaky import flaky
from pytest import approx
from scipy import stats

from dowhy.gcm import BayesianGaussianMixtureDistribution, EmpiricalDistribution, ScipyDistribution


def test_bayesian_gaussian_mixture_distribution():
    test_data = np.array([[0, 0], [0, 0], [1, 2], [1, 2]])

    approximated_data_distribution_model = BayesianGaussianMixtureDistribution()
    approximated_data_distribution_model.fit(test_data)
    assert approximated_data_distribution_model.draw_samples(5).shape == (5, 2)


def test_bayesian_gaussian_mixture_distribution_runtime_error():
    approximated_data_distribution_model = BayesianGaussianMixtureDistribution()
    with pytest.raises(RuntimeError):
        approximated_data_distribution_model.draw_samples(5)


def test_scipy_fixed_parametric_distribution():
    distribution = ScipyDistribution(stats.norm, loc=0, scale=1)

    assert distribution.parameters["loc"] == 0
    assert distribution.parameters["scale"] == 1


@flaky(max_runs=5)
def test_scipy_fittable_parametric_distribution():
    distribution = ScipyDistribution(stats.norm)

    X = np.random.normal(0, 1, 1000)
    distribution.fit(X)

    assert distribution.parameters["loc"] == approx(0, abs=0.1)
    assert distribution.parameters["scale"] == approx(1, abs=0.1)


@flaky(max_runs=5)
def test_scipy_auto_select_continuous_parametric_distribution():
    distribution = ScipyDistribution()

    X = np.random.normal(0, 1, 1000)
    distribution.fit(X)

    assert np.mean(distribution.draw_samples(1000)) == approx(0, abs=0.1)
    assert np.std(distribution.draw_samples(1000)) == approx(1, abs=0.1)


def test_empirical_distribution():
    X = np.random.normal(0, 1, 1000)

    distribution = EmpiricalDistribution()
    distribution.fit(X)

    X = list(X)

    for val in distribution.draw_samples(1000):
        assert val in X


@flaky(max_runs=5)
def test_fitted_parameters_assigned_correctly_using_normal_distribution():
    distribution = ScipyDistribution(stats.norm)
    distribution.fit(ScipyDistribution(stats.norm, loc=3, scale=2).draw_samples(10000))

    assert distribution.parameters["loc"] == approx(3, abs=0.3)
    assert distribution.parameters["scale"] == approx(2, abs=0.3)


@flaky(max_runs=5)
def test_fitted_parameters_assigned_correctly_using_beta_distribution():
    distribution = ScipyDistribution(stats.beta)
    distribution.fit(ScipyDistribution(stats.beta, a=2, b=0.5).draw_samples(10000))

    assert distribution.parameters["loc"] == approx(0, abs=0.1)
    assert distribution.parameters["scale"] == approx(1, abs=0.1)
    assert distribution.parameters["a"] == approx(2, abs=0.5)
    assert distribution.parameters["b"] == approx(0.5, abs=0.5)



from abc import ABC, abstractmethod

import numpy as np


class DensityEstimator(ABC):
    @abstractmethod
    def fit(self, X: np.ndarray) -> None:
        raise NotImplementedError

    @abstractmethod
    def density(self, X: np.ndarray) -> np.ndarray:
        """Returns the density of each input."""
        raise NotImplementedError



import numpy as np

from dowhy.do_sampler import DoSampler
from dowhy.utils.propensity_score import propensity_of_treatment_score, state_propensity_score


class WeightingSampler(DoSampler):
    def __init__(
        self,
        data,
        *args,
        params=None,
        variable_types=None,
        num_cores=1,
        keep_original_treatment=False,
        causal_model=None,
        **kwargs,
    ):
        """
        g, df, data_types

        """
        super().__init__(
            data,
            params=params,
            variable_types=variable_types,
            num_cores=num_cores,
            keep_original_treatment=keep_original_treatment,
            causal_model=causal_model,
        )

        self.logger.info("Using WeightingSampler for do sampling.")
        self.logger.info("Caution: do samplers assume iid data.")
        self.point_sampler = False

    def make_treatment_effective(self, x):
        to_sample = self._df.copy()
        if not self.keep_original_treatment:
            for treatment, value in x.items():
                to_sample = to_sample[to_sample[treatment] == value]
        self._df = to_sample

    def disrupt_causes(self):
        self._df["propensity_score"] = state_propensity_score(
            self._data,
            self._target_estimand.get_backdoor_variables(),
            self._treatment_names,
            variable_types=self._variable_types,
        )
        self._df["weight"] = self.compute_weights()

    def sample(self):
        self._df = self._df.sample(len(self._data), replace=True, weights=self._df["weight"])
        self._df.index = self._data.index

    def compute_weights(self):
        weights = 1.0 / self._df["propensity_score"]
        return weights



import os
import tempfile
from copy import deepcopy
from typing import Any, Dict, Optional, Tuple

import networkx as nx
import numpy as np
import pygraphviz
from matplotlib import image, pyplot


def _plot_causal_graph_graphviz(
    causal_graph: nx.Graph,
    display_causal_strengths: bool = True,
    causal_strengths: Optional[Dict[Tuple[Any, Any], float]] = None,
    filename: Optional[str] = None,
    display_plot: bool = True,
    figure_size: Optional[Tuple[int, int]] = None,
) -> None:
    if causal_strengths is None:
        causal_strengths = {}
    else:
        causal_strengths = deepcopy(causal_strengths)

    max_strength = 0.0
    for (source, target, strength) in causal_graph.edges(data="CAUSAL_STRENGTH", default=None):
        if (source, target) not in causal_strengths:
            causal_strengths[(source, target)] = strength
        if causal_strengths[(source, target)] is not None:
            max_strength = max(max_strength, abs(causal_strengths[(source, target)]))

    pygraphviz_graph = pygraphviz.AGraph(directed=isinstance(causal_graph, nx.DiGraph))

    for node in causal_graph.nodes:
        pygraphviz_graph.add_node(node)

    for (source, target) in causal_graph.edges():
        causal_strength = causal_strengths[(source, target)]
        if causal_strength is not None:
            if np.isinf(causal_strength):
                causal_strength = 10000
                tmp_label = "Inf"
            else:
                tmp_label = str(" %s" % str(int(causal_strength * 100) / 100))

            pygraphviz_graph.add_edge(
                str(source),
                str(target),
                label=tmp_label if display_causal_strengths else None,
                penwidth=str(_calc_arrow_width(causal_strength, max_strength)),
            )
        else:
            pygraphviz_graph.add_edge(str(source), str(target))

    pygraphviz_graph.layout(prog="dot")
    if filename is not None:
        filename, file_extension = os.path.splitext(filename)
        if file_extension == "":
            file_extension = ".pdf"
        pygraphviz_graph.draw(filename + file_extension)

    if display_plot:
        _plot_as_pyplot_figure(pygraphviz_graph, figure_size)


def _calc_arrow_width(strength: float, max_strength: float):
    return 0.1 + 4.0 * float(abs(strength)) / float(max_strength)


def _plot_as_pyplot_figure(pygraphviz_graph: pygraphviz.AGraph, figure_size: Optional[Tuple[int, int]] = None) -> None:
    with tempfile.TemporaryDirectory() as tmp_dir_name:
        pygraphviz_graph.draw(tmp_dir_name + os.sep + "Graph.png")
        img = image.imread(tmp_dir_name + os.sep + "Graph.png")

        if figure_size is not None:
            org_fig_size = pyplot.rcParams["figure.figsize"]
            pyplot.rcParams["figure.figsize"] = figure_size

        pyplot.imshow(img)
        pyplot.axis("off")
        pyplot.show()

        if figure_size is not None:
            pyplot.rcParams["figure.figsize"] = org_fig_size



import random

import numpy as np
import pytest
from flaky import flaky
from pytest import approx

from dowhy.gcm.shapley import ShapleyApproximationMethods, ShapleyConfig, estimate_shapley_values
from dowhy.gcm.stats import permute_features
from dowhy.gcm.util.general import means_difference


@pytest.fixture
def preserve_random_generator_state():
    numpy_state = np.random.get_state()
    random_state = random.getstate()
    yield
    np.random.set_state(numpy_state)
    random.setstate(random_state)


def test_given_few_features_when_estimate_shapley_values_with_auto_approx_then_returns_correct_result():
    X, coefficients = _generate_data(4)

    def model(x):
        return np.sum(coefficients * x, axis=1)

    shapley_values = estimate_shapley_values(
        lambda subset: _set_function_for_aggregated_feature_attribution(subset, X, model),
        X.shape[1],
        ShapleyConfig(approximation_method=ShapleyApproximationMethods.AUTO, n_jobs=1),
    )

    assert coefficients * (X[0, :] - np.mean(X, axis=0)) == approx(shapley_values, abs=0.001)


def test_given_many_features_when_estimate_shapley_values_with_auto_approx_then_returns_correct_result():
    X, coefficients = _generate_data(15)

    def model(x):
        return np.sum(coefficients * x, axis=1)

    shapley_values = estimate_shapley_values(
        lambda subset: _set_function_for_aggregated_feature_attribution(subset, X, model),
        X.shape[1],
        ShapleyConfig(approximation_method=ShapleyApproximationMethods.AUTO, n_jobs=1),
    )

    assert coefficients * (X[0, :] - np.mean(X, axis=0)) == approx(shapley_values, abs=0.001)


def test_estimate_shapley_values_symmetry_exact():
    X, coefficients = _generate_data(15)

    def model(x):
        return np.sum(coefficients * x, axis=1)

    shapley_values = estimate_shapley_values(
        lambda subset: _set_function_for_aggregated_feature_attribution(subset, X, model),
        X.shape[1],
        ShapleyConfig(approximation_method=ShapleyApproximationMethods.EXACT, n_jobs=1),
    )

    assert coefficients * (X[0, :] - np.mean(X, axis=0)) == approx(shapley_values, abs=0.001)


def test_estimate_shapley_values_symmetry_exact_fast():
    X, coefficients = _generate_data(15)

    def model(x):
        return np.sum(coefficients * x, axis=1)

    shapley_values = estimate_shapley_values(
        lambda subset: _set_function_for_aggregated_feature_attribution(subset, X, model),
        X.shape[1],
        ShapleyConfig(approximation_method=ShapleyApproximationMethods.EXACT_FAST, n_jobs=1),
    )

    assert coefficients * (X[0, :] - np.mean(X, axis=0)) == approx(shapley_values, abs=0.001)


def test_estimate_shapley_values_symmetry_approximation_via_subset():
    X, coefficients = _generate_data(15)

    def model(x):
        return np.sum(coefficients * x, axis=1)

    shapley_values = estimate_shapley_values(
        lambda subset: _set_function_for_aggregated_feature_attribution(subset, X, model),
        X.shape[1],
        ShapleyConfig(approximation_method=ShapleyApproximationMethods.SUBSET_SAMPLING),
    )

    assert coefficients * (X[0, :] - np.mean(X, axis=0)) == approx(shapley_values, abs=0.001)


def test_evaluate_set_function_via_shapley_symmetry_approximation_via_permutation():
    X, coefficients = _generate_data(15)

    def model(x):
        return np.sum(coefficients * x, axis=1)

    shapley_values = estimate_shapley_values(
        lambda subset: _set_function_for_aggregated_feature_attribution(subset, X, model),
        X.shape[1],
        ShapleyConfig(approximation_method=ShapleyApproximationMethods.PERMUTATION),
    )

    assert coefficients * (X[0, :] - np.mean(X, axis=0)) == approx(shapley_values, abs=0.001)


def test_estimate_shapley_values_symmetry_approximation_via_early_stopping():
    X, coefficients = _generate_data(15)

    def model(x):
        return np.sum(coefficients * x, axis=1)

    shapley_values = estimate_shapley_values(
        lambda subset: _set_function_for_aggregated_feature_attribution(subset, X, model),
        X.shape[1],
        ShapleyConfig(approximation_method=ShapleyApproximationMethods.EARLY_STOPPING),
    )

    assert coefficients * (X[0, :] - np.mean(X, axis=0)) == approx(shapley_values, abs=0.001)


def test_estimate_shapley_values_symmetry_approximation_via_subset_with_random_seed(preserve_random_generator_state):
    X, coefficients = _generate_data(15)

    def model(x):
        return np.sum(coefficients * x, axis=1)

    shapley_config = ShapleyConfig(approximation_method=ShapleyApproximationMethods.SUBSET_SAMPLING)
    assert estimate_shapley_values(
        lambda subset: _set_function_for_aggregated_feature_attribution(subset, X, model), X.shape[1], shapley_config
    ) != approx(
        estimate_shapley_values(
            lambda subset: _set_function_for_aggregated_feature_attribution(subset, X, model),
            X.shape[1],
            shapley_config,
        ),
        abs=0,
    )

    np.random.seed(0)
    shapley_values_1 = estimate_shapley_values(
        lambda subset: _set_function_for_aggregated_feature_attribution(subset, X, model), X.shape[1], shapley_config
    )
    np.random.seed(0)
    shapley_values_2 = estimate_shapley_values(
        lambda subset: _set_function_for_aggregated_feature_attribution(subset, X, model), X.shape[1], shapley_config
    )

    assert shapley_values_1 == approx(shapley_values_2, abs=0)


@flaky(max_runs=2)
def test_estimate_shapley_values_symmetry_approximation_via_permutation_with_random_seed(
    preserve_random_generator_state,
):
    X, coefficients = _generate_data(15)

    def model(x):
        return np.sum(coefficients * x, axis=1)

    shapley_config = ShapleyConfig(approximation_method=ShapleyApproximationMethods.PERMUTATION)
    assert estimate_shapley_values(
        lambda subset: _set_function_for_aggregated_feature_attribution(subset, X, model), X.shape[1], shapley_config
    ) != approx(
        estimate_shapley_values(
            lambda subset: _set_function_for_aggregated_feature_attribution(subset, X, model),
            X.shape[1],
            shapley_config,
        ),
        abs=0,
    )

    np.random.seed(0)
    shapley_values_1 = estimate_shapley_values(
        lambda subset: _set_function_for_aggregated_feature_attribution(subset, X, model), X.shape[1], shapley_config
    )
    np.random.seed(0)
    shapley_values_2 = estimate_shapley_values(
        lambda subset: _set_function_for_aggregated_feature_attribution(subset, X, model), X.shape[1], shapley_config
    )

    assert shapley_values_1 == approx(shapley_values_2, abs=0)


def test_estimate_shapley_values_symmetry_approximation_via_early_stopping_with_random_seed(
    preserve_random_generator_state,
):
    X, coefficients = _generate_data(15)

    def model(x):
        return np.sum(coefficients * x, axis=1)

    shapley_config = ShapleyConfig(approximation_method=ShapleyApproximationMethods.EARLY_STOPPING)
    assert estimate_shapley_values(
        lambda subset: _set_function_for_aggregated_feature_attribution(subset, X, model), X.shape[1], shapley_config
    ) != approx(
        estimate_shapley_values(
            lambda subset: _set_function_for_aggregated_feature_attribution(subset, X, model),
            X.shape[1],
            shapley_config,
        ),
        abs=0,
    )

    np.random.seed(0)
    shapley_values_1 = estimate_shapley_values(
        lambda subset: _set_function_for_aggregated_feature_attribution(subset, X, model), X.shape[1], shapley_config
    )
    np.random.seed(0)
    shapley_values_2 = estimate_shapley_values(
        lambda subset: _set_function_for_aggregated_feature_attribution(subset, X, model), X.shape[1], shapley_config
    )

    assert shapley_values_1 == approx(shapley_values_2, abs=0)


def _generate_data(num_vars):
    return np.random.normal(0, 1, (1000, num_vars)), np.random.choice(20, num_vars) - 10


def _set_function_for_aggregated_feature_attribution(subset, X, model):
    tmp = permute_features(X, np.arange(0, X.shape[1])[subset == 0], False)
    tmp[:, subset == 1] = X[0, subset == 1]

    return means_difference(model(tmp), X[0])



"""This module contains implementations of different anomaly scorers.

Classes and functions in this module should be considered experimental, meaning there might be breaking API changes in
the future.
"""

from typing import Optional

import numpy as np
from statsmodels.robust import mad

from dowhy.gcm.anomaly_scorer import AnomalyScorer
from dowhy.gcm.constant import EPS
from dowhy.gcm.density_estimator import DensityEstimator
from dowhy.gcm.util.general import shape_into_2d


class MedianCDFQuantileScorer(AnomalyScorer):
    """Given an anomalous observation x and samples from the distribution of X, this score represents:
        score(x) = 1 - 2 * min[P(X >= x), P(X <= x)]

    It scores the observation based on the quantile of x with respect to the distribution of X. Here, if the
    sample x lies in the tail of the distribution, we want to have a large score. Since we apriori don't know
    whether the sample falls on the left or right side of the median of X, we estimate the quantile on both
    sides and take the minimum. Here, these probabilities are estimated by counting and since half of the samples are
    on one side from the median, we need to multiply this by a factor of two to obtain the two-sided quantile.
    For example:
        X = [-3, -2, -1, 0, 1, 2, 3]
        x = 2.5
    Then, x falls in the right sided-quantile and only one sample in X is larger than x. Therefore, we get
        p(X >= x) = 1 / 7
        P(X <= x) = 6 / 7
    With the end score of:
       1 - 2 * min[P(X >= x), P(X <= x)] = 1 - 2 / 7 = 0.71

    Note: For equal samples, we contribute half of the count to the left and half of the count the right side.

    The higher the score, the less likely the sample comes from the distribution of X.
    """

    def __init__(self):
        self._distribution_samples = None

    def fit(self, X: np.ndarray) -> None:
        if (X.ndim == 2 and X.shape[1] > 1) or X.ndim > 2:
            raise ValueError("The MedianCDFQuantileScorer currently only supports one-dimensional data!")

        self._distribution_samples = X.reshape(-1)

    def score(self, X: np.ndarray) -> np.ndarray:
        if self._distribution_samples is None:
            raise ValueError("Scorer has not been fitted!")

        X = shape_into_2d(X)

        equal_samples = np.sum(X == self._distribution_samples, axis=1)
        greater_samples = np.sum(X > self._distribution_samples, axis=1) + equal_samples / 2
        smaller_samples = np.sum(X < self._distribution_samples, axis=1) + equal_samples / 2

        return (
            1 - 2 * np.amin(np.vstack([greater_samples, smaller_samples]), axis=0) / self._distribution_samples.shape[0]
        )


class RescaledMedianCDFQuantileScorer(AnomalyScorer):
    """Given an anomalous observation x and samples from the distribution of X, this score represents:
        score(x) = -log(2 * min[P(X >= x), P(X <= x)])

    This is a rescaled version of the score s obtained by the :class:`~dowhy.gcm.anomaly_scorers.MedianCDFQuantileScorer`
    by calculating the negative log-probability -log(1 - s). This has the advantage that small differences in the
    probabilities are amplified, especially when they are close to 0. For instance, the difference between
    probabilities 0.02 and 0.01 seems to be small and insignificant, but the rescaled difference would be significantly
    larger: -log(0.02) - log(0.01) −= 8.5

    The higher the score, the less likely the sample comes from the distribution of X.
    """

    def __init__(self):
        self._scorer = MedianCDFQuantileScorer()

    def fit(self, X: np.ndarray) -> None:
        self._scorer.fit(X)

    def score(self, X: np.ndarray) -> np.ndarray:
        scores = 1 - self._scorer.score(X)
        scores[scores == 0] = EPS

        return -np.log(scores)


class ITAnomalyScorer(AnomalyScorer):
    """Transforms any anomaly scorer into an information theoretic (IT) score. This means, given a scorer S(x), an
    anomalous observation x and samples from the distribution of X, this scorer class represents:
        score(x) = -log(P(S(X) >= S(x)))

    This is, the negative logarithm of the probability to get the same or a higher score with (random) samples from X
    compared to the score obtained based on the anomalous observation x. By this, the score of arbitrarily different
    anomaly scorers become comparable information theoretic quantities. The new score -log(P(S(X) >= S(x))) can also
    be seen as "The higher the score, the rarer the anomaly event". For instance, if we have S(x) = c, but observe
    the same or higher scores in 50% or even 100% of all samples in X, then this is not really a rare event, and thus,
    not an anomaly. As mentioned above, transforming it into an IT score makes arbitrarily different anomaly scorer
    with potentially completely different scaling comparable. For example, one could compare the IT score of
    isolation forests with z-scores.

    For more details about IT scores, see:
        Causal structure based root cause analysis of outliers
        Kailash Budhathoki, Patrick Bloebaum, Lenon Minorics, Dominik Janzing (2022)

    The higher the score, the higher the likelihood that the observations is an anomaly.
    """

    def __init__(self, anomaly_scorer: AnomalyScorer):
        self._anomaly_scorer = anomaly_scorer
        self._distribution_samples = None
        self._scores_of_distribution_samples = None

    def fit(self, X: np.ndarray) -> None:
        self._distribution_samples = shape_into_2d(X)
        self._anomaly_scorer.fit(self._distribution_samples)
        self._scores_of_distribution_samples = self._anomaly_scorer.score(self._distribution_samples).reshape(-1)

    def score(self, X: np.ndarray) -> np.ndarray:
        X = shape_into_2d(X)
        scores_of_samples_to_score = self._anomaly_scorer.score(X).reshape(-1, 1)
        return -np.log(
            (np.sum(self._scores_of_distribution_samples >= scores_of_samples_to_score, axis=1) + 0.5)
            / (self._scores_of_distribution_samples.shape[0] + 0.5)
        )


class MeanDeviationScorer(AnomalyScorer):
    """Given an anomalous observation x and samples from the distribution of X, this score represents:
        score(x) = |x - E[X]| / std[X]

    This scores the given sample based on its distance to the mean of X and scaled by the standard deviation of X. This
    is also equivalent to the Z-score in Gaussian variables.

    The higher the score, the higher the deviation of the observation from the mean of X.
    """

    def __init__(self):
        self._mean = None
        self._std = None

    def fit(self, X: np.ndarray) -> None:
        self._mean = np.mean(X)
        self._std = np.std(X)

    def score(self, X: np.ndarray) -> np.ndarray:
        if self._mean is None or self._std is None:
            raise ValueError("Scorer has not been fitted!")

        return abs(X - self._mean) / self._std


class MedianDeviationScorer(AnomalyScorer):
    """Given an anomalous observation x and samples from the distribution of X, this score represents:
        score(x) = |x - med[X]| / mad[X]

    This scores the given sample based on its distance to the median of X and scaled by the median absolute deviation
    of X.

    The higher the score, the higher the deviation of the observation from the median of X.
    """

    def __init__(self):
        self._median = None
        self._mad = None

    def fit(self, X: np.ndarray) -> None:
        self._median = np.median(X)
        self._mad = mad(X)

    def score(self, X: np.ndarray) -> np.ndarray:
        if self._median is None or self._mad is None:
            raise ValueError("Scorer has not been fitted!")

        return abs(X - self._median) / self._mad


class InverseDensityScorer(AnomalyScorer):
    """Estimates an anomaly score based on 1 / p(x), where x is the data to score. The density value p(x) is estimated
    using the given density estimator. If None is given, a Gaussian mixture model is used by default.

    Note: The given density estimator needs to support the data types, i.e. if the data has categorical values, the
    density estimator needs to be able to handle that. The default Gaussian model can only handle numeric data.

    Note: If the density p(x) is 0, a nan or inf could be returned.
    """

    def __init__(self, density_estimator: Optional[DensityEstimator] = None):
        if density_estimator is None:
            from dowhy.gcm.density_estimators import GaussianMixtureDensityEstimator

            density_estimator = GaussianMixtureDensityEstimator()
        self._density_estimator = density_estimator
        self._fitted = False

    def fit(self, X: np.ndarray) -> None:
        self._density_estimator.fit(X)
        self._fitted = True

    def score(self, X: np.ndarray) -> np.ndarray:
        if not self._fitted:
            raise ValueError("Scorer has not been fitted!")

        return 1 / self._density_estimator.density(X)



import numpy as np
from flaky import flaky
from pytest import approx
from sklearn.linear_model import LogisticRegression

from dowhy.gcm.ml import SklearnClassificationModel, create_linear_regressor, create_logistic_regression_classifier


@flaky(max_runs=5)
def test_categorical_features():
    X0 = np.random.normal(0, 1, 1000)
    X1 = np.random.choice(3, 1000).astype(str)
    X2 = []

    for i in range(1000):
        tmp_value = 2 * X0[i]

        if X1[i] == "0":
            tmp_value -= 5
        elif X1[i] == "1":
            tmp_value += 10
        else:
            tmp_value += 5

        X2.append(tmp_value)

    X2 = np.array(X2)

    inputs = np.array([X0, X1], dtype=object).T

    mdl = create_linear_regressor()
    mdl.fit(inputs, X2)

    assert mdl.predict(np.array([[2, "1"]], dtype=object)) == approx(14)
    assert mdl.predict(inputs) == approx(X2.reshape(-1, 1))


def test_categorical_inputs():
    X0 = np.random.normal(0, 1, 1000)
    X1 = np.random.choice(3, 1000).astype(str)
    X2 = []

    for i in range(1000):
        tmp_value = 2 * X0[i]

        if X1[i] == "0":
            tmp_value -= 5
        elif X1[i] == "1":
            tmp_value += 10
        else:
            tmp_value += 5

        X2.append(tmp_value)

    X2 = (X2 > np.median(X2)).astype(str)

    inputs = np.array([X0, X1], dtype=object).T

    mdl = create_logistic_regression_classifier()
    mdl.fit(inputs, X2)

    X2["True" == 1] = 1
    X2["False" == 0] = 0

    assert mdl.predict_probabilities(np.array([[2, "1"]], dtype=object)) == approx(np.array([[0, 1]]), abs=0.01)
    assert np.sum(np.argmax(mdl.predict_probabilities(inputs), axis=1) != X2) < 20

    _, counts = np.unique(mdl.predict(inputs), return_counts=True)
    assert counts / 1000 == approx(np.array([0.5, 0.5]), abs=0.05)


def test_when_cloning_sklearn_classification_model_then_returns_a_cloned_object():
    logistic_regression_model = LogisticRegression()
    mdl = SklearnClassificationModel(logistic_regression_model)
    cloned_mdl = mdl.clone()

    assert isinstance(cloned_mdl, SklearnClassificationModel)
    assert isinstance(cloned_mdl.sklearn_model, LogisticRegression)
    assert mdl != cloned_mdl
    assert cloned_mdl.sklearn_model != logistic_regression_model



import numpy as np
from flaky import flaky

from dowhy.gcm.ml import create_polynom_logistic_regression_classifier


@flaky(max_runs=3)
def test_when_fit_and_predict_polynom_classifier_then_returns_accurate_results():
    def _generate_data():
        X = np.random.normal(0, 1, (1000, 2))
        Y = []

        for x in X:
            if x[0] * x[1] > 0:
                Y.append("Class 0")
            else:
                Y.append("Class 1")

        return X, np.array(Y)

    X_training, Y_training = _generate_data()
    X_test, Y_test = _generate_data()
    mdl = create_polynom_logistic_regression_classifier()
    mdl.fit(X_training, Y_training)

    assert np.sum(mdl.predict(X_test).reshape(-1) == Y_test) > 950



import networkx as nx
import numpy as np
import pandas as pd

from dowhy.gcm.util import plot, plot_adjacency_matrix


def test_when_plot_does_not_raise_exception():
    plot(nx.DiGraph([("X", "Y"), ("Y", "Z")]))


def test_plot_adjacency_matrix():
    causal_graph = pd.DataFrame(
        np.array([[0, 1, 0], [0, 0, 1], [0, 0, 0]]), columns=["X", "Y", "Z"], index=["X", "Y", "Z"]
    )

    # Check if calling the method causes some import or runtime errors
    plot_adjacency_matrix(causal_graph, is_directed=True)
    # TODO: Plotting undirected graphs with networkx causes an error when an older networkx version is used with a newer
    #  newer matplotlib version:
    #  AttributeError: module 'matplotlib.cbook' has no attribute 'is_numlike'
    #  Networkx 2.4+ should fix this issue.
    # plot_adjacency_matrix(causal_graph, is_directed=False)


def test_given_causal_strengths_when_plot_graph_then_does_not_modify_input_object():
    causal_strength = {("X", "Y"): 10}

    plot(nx.DiGraph([("X", "Y"), ("Y", "Z")]), causal_strengths=causal_strength)

    assert causal_strength == {("X", "Y"): 10}



import itertools

import dowhy.datasets
from dowhy import CausalModel


class TestEstimator(object):
    def __init__(self, error_tolerance, Estimator, identifier_method="backdoor"):
        print("Error tolerance is", error_tolerance)
        self._error_tolerance = error_tolerance
        self._Estimator = Estimator
        self._identifier_method = identifier_method

    def average_treatment_effect_test(
        self,
        dataset="linear",
        beta=10,
        num_common_causes=1,
        num_instruments=1,
        num_effect_modifiers=0,
        num_treatments=1,
        num_frontdoor_variables=0,
        num_samples=100000,
        treatment_is_binary=True,
        treatment_is_category=False,
        outcome_is_binary=False,
        confidence_intervals=False,
        test_significance=False,
        method_params=None,
    ):
        if dataset == "linear":
            data = dowhy.datasets.linear_dataset(
                beta=beta,
                num_common_causes=num_common_causes,
                num_instruments=num_instruments,
                num_effect_modifiers=num_effect_modifiers,
                num_treatments=num_treatments,
                num_frontdoor_variables=num_frontdoor_variables,
                num_samples=num_samples,
                treatment_is_binary=treatment_is_binary,
                treatment_is_category=treatment_is_category,
                outcome_is_binary=outcome_is_binary,
            )
        elif dataset == "simple-iv":
            data = dowhy.datasets.simple_iv_dataset(
                beta=beta,
                num_treatments=num_treatments,
                num_samples=num_samples,
                treatment_is_binary=treatment_is_binary,
                outcome_is_binary=outcome_is_binary,
            )
        else:
            raise ValueError("Dataset type not supported.")

        model = CausalModel(
            data=data["df"],
            treatment=data["treatment_name"],
            outcome=data["outcome_name"],
            graph=data["gml_graph"],
            proceed_when_unidentifiable=True,
            test_significance=test_significance,
        )
        target_estimand = model.identify_effect()
        target_estimand.set_identifier_method(self._identifier_method)
        estimator_ate = self._Estimator(
            identified_estimand=target_estimand,
            **method_params,
        )

        estimator_ate.fit(
            data["df"],
            data["treatment_name"],
            data["outcome_name"],
            effect_modifier_names=data["effect_modifier_names"],
        )

        true_ate = data["ate"]
        ate_estimate = estimator_ate.estimate_effect(
            control_value=0,
            treatment_value=1,
            test_significance=test_significance,
            evaluate_effect_strength=False,
            confidence_intervals=confidence_intervals,
            target_units="ate",
            **method_params,
        )
        str(ate_estimate)  # checking if str output is correctly created
        error = abs(ate_estimate.value - true_ate)
        print(
            "Error in ATE estimate = {0} with tolerance {1}%. Estimated={2},True={3}".format(
                error, self._error_tolerance * 100, ate_estimate.value, true_ate
            )
        )
        res = True if (error < abs(true_ate) * self._error_tolerance) else False
        assert res
        # Compute confidence intervals, standard error and significance tests
        if confidence_intervals:
            ate_estimate.get_confidence_intervals()
            ate_estimate.get_confidence_intervals(confidence_level=0.99)
            ate_estimate.get_confidence_intervals(method="bootstrap")
            ate_estimate.get_standard_error()
            ate_estimate.get_standard_error(method="bootstrap")
        if test_significance:
            ate_estimate.test_stat_significance()
            ate_estimate.test_stat_significance(method="bootstrap")

    def average_treatment_effect_testsuite(
        self,
        tests_to_run="all",
        num_common_causes=[2, 3],
        num_instruments=[
            1,
        ],
        num_effect_modifiers=[
            0,
        ],
        num_treatments=[
            1,
        ],
        num_frontdoor_variables=[
            0,
        ],
        treatment_is_binary=[
            True,
        ],
        treatment_is_category=[
            False,
        ],
        outcome_is_binary=[
            False,
        ],
        confidence_intervals=[
            False,
        ],
        test_significance=[
            False,
        ],
        dataset="linear",
        method_params=None,
    ):
        args_dict = {
            "num_common_causes": num_common_causes,
            "num_instruments": num_instruments,
            "num_effect_modifiers": num_effect_modifiers,
            "num_treatments": num_treatments,
            "num_frontdoor_variables": num_frontdoor_variables,
            "treatment_is_binary": treatment_is_binary,
            "treatment_is_category": treatment_is_category,
            "outcome_is_binary": outcome_is_binary,
            "confidence_intervals": confidence_intervals,
            "test_significance": test_significance,
        }
        keys, values = zip(*args_dict.items())
        configs = [dict(zip(keys, v)) for v in itertools.product(*values)]
        for cfg in configs:
            print("\nConfig:", cfg)
            cfg["dataset"] = dataset
            cfg["method_params"] = method_params
            self.average_treatment_effect_test(**cfg)

    def custom_data_average_treatment_effect_test(self, data):
        model = CausalModel(
            data=data["df"],
            treatment=data["treatment_name"],
            outcome=data["outcome_name"],
            graph=data["gml_graph"],
            proceed_when_unidentifiable=True,
            test_significance=None,
        )
        target_estimand = model.identify_effect()
        estimator_ate = self._Estimator(
            identified_estimand=target_estimand,
            test_significance=None,
        )
        estimator_ate.fit(
            data["df"],
            treatment=data["treatment_name"],
            outcome=data["outcome_name"],
        )
        true_ate = data["ate"]
        ate_estimate = estimator_ate.estimate_effect()
        error = ate_estimate.value - true_ate
        print(
            "Error in ATE estimate = {0} with tolerance {1}%. Estimated={2},True={3}".format(
                error, self._error_tolerance * 100, ate_estimate.value, true_ate
            )
        )
        res = True if (error < true_ate * self._error_tolerance) else False
        assert res



import pytest

from dowhy.causal_graph import CausalGraph

from .example_graphs import TEST_GRAPH_SOLUTIONS


class IdentificationTestGraphSolution(object):
    def __init__(
        self,
        graph_str,
        observed_variables,
        biased_sets,
        minimal_adjustment_sets,
        maximal_adjustment_sets,
        direct_maximal_adjustment_sets=None,
    ):
        self.graph = CausalGraph("X", "Y", graph_str, observed_node_names=observed_variables)
        self.graph_str = graph_str
        self.observed_variables = observed_variables
        self.biased_sets = biased_sets
        self.minimal_adjustment_sets = minimal_adjustment_sets
        self.maximal_adjustment_sets = maximal_adjustment_sets
        self.direct_maximal_adjustment_sets = (
            maximal_adjustment_sets if direct_maximal_adjustment_sets is None else direct_maximal_adjustment_sets
        )


@pytest.fixture(params=TEST_GRAPH_SOLUTIONS.keys())
def example_graph_solution(request):
    return IdentificationTestGraphSolution(**TEST_GRAPH_SOLUTIONS[request.param])



"""This module defines the fundamental interfaces and functions related to causal graphs in graphical causal models.

Classes and functions in this module should be considered experimental, meaning there might be breaking API changes in
the future.
"""

from abc import ABC, abstractmethod
from typing import Any, List

import networkx as nx
import numpy as np
from networkx.algorithms.dag import has_cycle
from typing_extensions import Protocol

# This constant is used as key when storing/accessing models as causal mechanisms in graph node attributes
CAUSAL_MECHANISM = "causal_mechanism"

# This constant is used as key when storing the parents of a node during fitting. It's used for validation purposes
# afterwards.
PARENTS_DURING_FIT = "parents_during_fit"


class HasNodes(Protocol):
    """This protocol defines a trait for classes having nodes."""

    @property
    @abstractmethod
    def nodes(self):
        """:returns Dict[Any, Dict[Any, Any]]"""
        raise NotImplementedError


class HasEdges(Protocol):
    """This protocol defines a trait for classes having edges."""

    @property
    @abstractmethod
    def edges(self):
        """:returns a Dict[Tuple[Any, Any], Dict[Any, Any]]"""
        raise NotImplementedError


class DirectedGraph(HasNodes, HasEdges, Protocol):
    """A protocol representing a directed graph as needed by graphical causal models.

    This protocol specifically defines a subset of the networkx.DiGraph class, which make that class automatically
    compatible with DirectedGraph. While in most cases a networkx.DiGraph is the class of choice when constructing
    a causal graph, anyone can choose to provide their own implementation of the DirectGraph interface.
    """

    @abstractmethod
    def predecessors(self, node):
        raise NotImplementedError


class StochasticModel(ABC):
    """A stochastic model represents a model used for causal mechanisms for root nodes in a graphical causal model."""

    @abstractmethod
    def fit(self, X: np.ndarray) -> None:
        """Fits the model according to the data."""
        raise NotImplementedError

    @abstractmethod
    def draw_samples(self, num_samples: int) -> np.ndarray:
        """Draws samples for the fitted model."""
        raise NotImplementedError

    @abstractmethod
    def clone(self):
        raise NotImplementedError


class ConditionalStochasticModel(ABC):
    """A conditional stochastic model represents a model used for causal mechanisms for non-root nodes in a graphical
    causal model."""

    @abstractmethod
    def fit(self, X: np.ndarray, Y: np.ndarray) -> None:
        """Fits the model according to the data."""
        raise NotImplementedError

    @abstractmethod
    def draw_samples(self, parent_samples: np.ndarray) -> np.ndarray:
        """Draws samples for the fitted model."""
        raise NotImplementedError

    @abstractmethod
    def clone(self):
        raise NotImplementedError


class FunctionalCausalModel(ConditionalStochasticModel):
    """Represents a Functional Causal Model (FCM), a specific type of conditional stochastic model, that is defined
    as:
        Y := f(X, N), N: Noise
    """

    def draw_samples(self, parent_samples: np.ndarray) -> np.ndarray:
        return self.evaluate(parent_samples, self.draw_noise_samples(parent_samples.shape[0]))

    @abstractmethod
    def draw_noise_samples(self, num_samples: int) -> np.ndarray:
        raise NotImplementedError

    @abstractmethod
    def evaluate(self, parent_samples: np.ndarray, noise_samples: np.ndarray) -> np.ndarray:
        raise NotImplementedError


class InvertibleFunctionalCausalModel(FunctionalCausalModel, ABC):
    @abstractmethod
    def estimate_noise(self, target_samples: np.ndarray, parent_samples: np.ndarray) -> np.ndarray:
        raise NotImplementedError


def is_root_node(causal_graph: DirectedGraph, node: Any) -> bool:
    return list(causal_graph.predecessors(node)) == []


def get_ordered_predecessors(causal_graph: DirectedGraph, node: Any) -> List[Any]:
    """This function returns predecessors of a node in a well-defined order.

    This is necessary, because we select subsets of columns in Dataframes by using a node's parents, and these parents
    might not be returned in a reliable order.
    """
    return sorted(causal_graph.predecessors(node))


def node_connected_subgraph_view(g: DirectedGraph, node: Any) -> Any:
    """Returns a view of the provided graph g that contains only nodes connected to the node passed in"""
    # can't use nx.node_connected_component, because it doesn't work with DiGraphs.
    # Hence a manual loop:
    return nx.induced_subgraph(g, [n for n in g.nodes if nx.has_path(g, n, node)])


def clone_causal_models(source: HasNodes, destination: HasNodes):
    for node in destination.nodes:
        if CAUSAL_MECHANISM in source.nodes[node]:
            destination.nodes[node][CAUSAL_MECHANISM] = source.nodes[node][CAUSAL_MECHANISM].clone()


def validate_acyclic(causal_graph: DirectedGraph) -> None:
    if has_cycle(causal_graph):
        raise RuntimeError("The graph contains a cycle, but an acyclic graph is expected!")


def validate_causal_dag(causal_graph: DirectedGraph) -> None:
    validate_acyclic(causal_graph)
    validate_causal_graph(causal_graph)


def validate_causal_graph(causal_graph: DirectedGraph) -> None:
    for node in causal_graph.nodes:
        validate_node(causal_graph, node)


def validate_node(causal_graph: DirectedGraph, node: Any) -> None:
    validate_causal_model_assignment(causal_graph, node)
    validate_local_structure(causal_graph, node)


def validate_causal_model_assignment(causal_graph: DirectedGraph, target_node: Any) -> None:
    validate_node_has_causal_model(causal_graph, target_node)

    causal_model = causal_graph.nodes[target_node][CAUSAL_MECHANISM]

    if is_root_node(causal_graph, target_node):
        if not isinstance(causal_model, StochasticModel):
            raise RuntimeError(
                "Node %s is a root node and, thus, requires a StochasticModel, "
                "but a %s was found!" % (target_node, causal_model)
            )
    elif not isinstance(causal_model, ConditionalStochasticModel):
        raise RuntimeError(
            "Node %s has parents and, thus, requires a ConditionalStochasticModel, "
            "but a %s was found!" % (target_node, causal_model)
        )


def validate_local_structure(causal_graph: DirectedGraph, node: Any) -> None:
    if PARENTS_DURING_FIT not in causal_graph.nodes[node] or causal_graph.nodes[node][
        PARENTS_DURING_FIT
    ] != get_ordered_predecessors(causal_graph, node):
        raise RuntimeError(
            "The causal mechanism of node %s is not fitted to the graphical structure! Fit all"
            "causal models in the graph first. If the mechanism is already fitted based on the causal"
            "parents, consider to update the persisted parents for that node manually." % node
        )


def validate_node_has_causal_model(causal_graph: HasNodes, node: Any) -> None:
    validate_node_in_graph(causal_graph, node)

    if CAUSAL_MECHANISM not in causal_graph.nodes[node]:
        raise ValueError("Node %s has no assigned causal mechanism!" % node)


def validate_node_in_graph(causal_graph: HasNodes, node: Any) -> None:
    if node not in causal_graph.nodes:
        raise ValueError("Node %s can not be found in the given graph!" % node)
